import ConfigParser
import StringIO
import sys
import warnings

import argparse
from lxml import etree
import msgpack

from turnstile import database
from turnstile import limits
from turnstile import tools

import tests
from tests.test_database import FakeDatabase


class FakeLimit(tests.GenericFakeClass):
    attrs = dict(
        int_attr=dict(
            type=int,
            default=1,
            ),
        float_attr=dict(
            type=float,
            default=1.5,
            ),
        str_attr=dict(
            type=str,
            default='spam',
            ),
        list_attr=dict(
            type=list,
            subtype=str,
            default=lambda: [],
            ),
        dict_attr=dict(
            type=dict,
            subtype=str,
            default=lambda: {},
            ),
        required=dict(
            type=str,
            ),
        )

    _limit_full_name = "FakeLimit"

    @classmethod
    def hydrate(cls, db, limit):
        return cls(db, **limit)

    def dehydrate(self):
        return self.kwargs

    def __getattr__(self, attr):
        return self.kwargs[attr]

    def __repr__(self):
        return repr(self.kwargs)


class FakeConfigParser(object):
    _cfg_options = dict(
        bad_config={},
        good_config=dict(
            connection=dict(
                host='example.com',
                ),
            ),
        alt_limits=dict(
            connection=dict(
                host='example.com',
                limits_key='alternate',
                ),
            ),
        alt_control=dict(
            connection=dict(
                host='example.com',
                control_channel='alternate',
                ),
            ),
        )

    def __init__(self):
        self.config = None

    def read(self, cfg_file):
        self.config = self._cfg_options[cfg_file]

    def has_section(self, section):
        return section in self.config

    def items(self, section):
        return self.config[section].items()


class FakeNamespace(object):
    config = 'config'
    limits_file = 'limits.xml'
    reload = False
    dry_run = True
    debug = False


class FakeArgumentParser(object):
    def __init__(self, *args, **kwargs):
        pass

    def add_argument(self, *args, **kwargs):
        pass

    def parse_args(self):
        return FakeNamespace()


class FakeArgumentParserDebug(FakeArgumentParser):
    def parse_args(self):
        ns = super(FakeArgumentParserDebug, self).parse_args()
        ns.debug = True
        return ns


class TestParseLimitNode(tests.TestCase):
    imports = {
        'FakeLimit': FakeLimit,
        }

    def test_parse_basic(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(required='spam'))

    def test_parse_missing_requirement(self):
        limit_xml = '<limit class="FakeLimit"></limit>'
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            with self.assertRaises(TypeError):
                limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

    def test_parse_unknown_elem(self):
        limit_xml = """<limit class="FakeLimit">
    <attribute name="badelem">spam</attribute>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 1)
            self.assertIn("Unrecognized element 'attribute' while parsing "
                          "limit at index 1; ignoring...", w[-1].message)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(required='spam'))

    def test_parse_unknown_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="unknown_attr">spam</attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 1)
            self.assertIn("Limit at index 1 does not accept an attribute "
                          "'unknown_attr'; ignoring...", w[-1].message)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(required='spam'))

    def test_parse_int_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="int_attr">5</attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                int_attr=5,
                ))

    def test_parse_float_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="float_attr">3.14</attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                float_attr=3.14,
                ))

    def test_parse_str_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="str_attr">ni</attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                str_attr='ni',
                ))

    def test_parse_list_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="list_attr">
        <value>spam</value>
        <value>ni</value>
    </attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                list_attr=['spam', 'ni'],
                ))

    def test_parse_list_attr_unknown_elem(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="list_attr">
        <val>spam</val>
        <value>ni</value>
    </attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 1)
            self.assertIn("Unrecognized element 'val' while parsing "
                          "'list_attr' attribute of limit at index 1; "
                          "ignoring...", w[-1].message)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                list_attr=['ni'],
                ))

    def test_parse_dict_attr(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="dict_attr">
        <value key="foo">spam</value>
        <value key="bar">ni</value>
    </attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 0)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                dict_attr=dict(foo='spam', bar='ni'),
                ))

    def test_parse_dict_attr_unknown_elem(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="dict_attr">
        <val key="foo">spam</val>
        <value key="bar">ni</value>
    </attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 1)
            self.assertIn("Unrecognized element 'val' while parsing "
                          "'dict_attr' attribute of limit at index 1; "
                          "ignoring...", w[-1].message)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                dict_attr=dict(bar='ni'),
                ))

    def test_parse_dict_attr_missing_key(self):
        limit_xml = """<limit class="FakeLimit">
    <attr name="dict_attr">
        <value>spam</value>
        <value key="bar">ni</value>
    </attr>
    <attr name="required">spam</attr>
</limit>"""
        limit_node = etree.fromstring(limit_xml)

        with warnings.catch_warnings(record=True) as w:
            limit = tools.parse_limit_node('db', 1, limit_node)

            self.assertEqual(len(w), 1)
            self.assertIn("Missing 'key' attribute of 'value' element "
                          "while parsing 'dict_attr' attribute of limit "
                          "at index 1; ignoring...", w[-1].message)

        self.assertIsInstance(limit, FakeLimit)
        self.assertEqual(limit.args, ('db',))
        self.assertEqual(limit.kwargs, dict(
                required='spam',
                dict_attr=dict(bar='ni'),
                ))


class TestMakeLimitNode(tests.TestCase):
    def setUp(self):
        super(TestMakeLimitNode, self).setUp()

        self.root = etree.Element('root')
        self.limit_kwargs = dict(
            int_attr=1,
            float_attr=1.5,
            str_attr='spam',
            list_attr=[],
            dict_attr={},
            required='required',
            )

    def _match_xml(self, expected):
        actual = etree.tostring(self.root)
        self.assertEqual(actual, '<root>%s</root>' % expected)

    def _prepare_limit(self, **kwargs):
        self.limit_kwargs.update(kwargs)
        return FakeLimit(**self.limit_kwargs)

    def test_make_basic(self):
        limit = self._prepare_limit()
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="required">required</attr>'
                        '</limit>')

    def test_make_int_attr(self):
        limit = self._prepare_limit(int_attr=5)
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="int_attr">5</attr>'
                        '<attr name="required">required</attr>'
                        '</limit>')

    def test_make_float_attr(self):
        limit = self._prepare_limit(float_attr=3.14)
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="float_attr">3.14</attr>'
                        '<attr name="required">required</attr>'
                        '</limit>')

    def test_make_str_attr(self):
        limit = self._prepare_limit(str_attr="nospam")
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="required">required</attr>'
                        '<attr name="str_attr">nospam</attr>'
                        '</limit>')

    def test_make_list_attr(self):
        limit = self._prepare_limit(list_attr=['spam', 'ni'])
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="list_attr">'
                        '<value>spam</value>'
                        '<value>ni</value>'
                        '</attr>'
                        '<attr name="required">required</attr>'
                        '</limit>')

    def test_make_dict_attr(self):
        limit = self._prepare_limit(dict_attr=dict(foo='spam', bar='ni'))
        tools.make_limit_node(self.root, limit)

        self._match_xml('<limit class="FakeLimit">'
                        '<attr name="dict_attr">'
                        '<value key="bar">ni</value>'
                        '<value key="foo">spam</value>'
                        '</attr>'
                        '<attr name="required">required</attr>'
                        '</limit>')


class ConsoleScriptsTestCase(tests.TestCase):
    def setUp(self):
        super(ConsoleScriptsTestCase, self).setUp()

        self.result = None
        self.subargs = None

        def fake_subroutine(*args, **kwargs):
            self.subargs = (args, kwargs)

            if isinstance(self.result, Exception):
                raise self.result
            return self.result

        self.stubs.Set(tools, '_%s' % self.subroutine.__name__,
                       fake_subroutine)

    def stub_argparse(self, debug=False):
        cls = FakeArgumentParserDebug if debug else FakeArgumentParser
        self.stubs.Set(argparse, 'ArgumentParser', cls)


class TestConsoleSetupLimits(ConsoleScriptsTestCase):
    subroutine = staticmethod(tools.setup_limits)

    def test_basic(self):
        self.stub_argparse()
        res = self.subroutine()

        self.assertEqual(res, None)
        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', False, True, False), {}))

    def test_exception(self):
        self.stub_argparse()
        self.result = Exception("An error occurred")
        res = self.subroutine()

        self.assertEqual(res, "An error occurred")
        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', False, True, False), {}))

    def test_exception_debug(self):
        self.stub_argparse(True)
        self.result = Exception("An error occurred")

        with self.assertRaises(Exception):
            res = self.subroutine()

        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', False, True, True), {}))


class TestConsoleDumpLimits(ConsoleScriptsTestCase):
    subroutine = staticmethod(tools.dump_limits)

    def test_basic(self):
        self.stub_argparse()
        res = self.subroutine()

        self.assertEqual(res, None)
        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', False), {}))

    def test_exception(self):
        self.stub_argparse()
        self.result = Exception("An error occurred")
        res = self.subroutine()

        self.assertEqual(res, "An error occurred")
        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', False), {}))

    def test_exception_debug(self):
        self.stub_argparse(True)
        self.result = Exception("An error occurred")

        with self.assertRaises(Exception):
            res = self.subroutine()

        self.assertEqual(self.subargs,
                         (('config', 'limits.xml', True), {}))


class TestParseConfig(tests.TestCase):
    def setUp(self):
        super(TestParseConfig, self).setUp()

        def fake_initialize(cfg):
            return cfg

        self.stubs.Set(database, 'initialize', fake_initialize)
        self.stubs.Set(ConfigParser, 'SafeConfigParser', FakeConfigParser)

    def test_basic(self):
        result = tools.parse_config('good_config')

        self.assertEqual(result,
                         (dict(host='example.com'), 'limits', 'control'))

    def test_alt_limits(self):
        result = tools.parse_config('alt_limits')

        self.assertEqual(result,
                         (dict(host='example.com'), 'alternate', 'control'))

    def test_alt_control(self):
        result = tools.parse_config('alt_control')

        self.assertEqual(result,
                         (dict(host='example.com'), 'limits', 'alternate'))

    def test_missing_connection(self):
        with self.assertRaises(Exception):
            result = tools.parse_config('bad_config')


class BaseToolTest(tests.TestCase):
    def setUp(self):
        super(BaseToolTest, self).setUp()

        self.fakedb = FakeDatabase()
        self.stderr = StringIO.StringIO()

        def fake_parse_config(config):
            self.assertEqual(config, 'config.file')

            return self.fakedb, 'limits', 'control'

        self.stubs.Set(tools, 'parse_config', fake_parse_config)
        self.stubs.Set(sys, 'stderr', self.stderr)


class TestToolSetupLimits(BaseToolTest):
    def setUp(self):
        super(TestToolSetupLimits, self).setUp()

        self.fail_idx = None
        self.parsed = []
        self.lims = None
        self.cmds = []

        def fake_parse_limit_node(db, idx, lim):
            lim_id = lim.get('id')

            self.parsed.append((idx, lim_id))
            if idx == self.fail_idx:
                raise Exception("Failed to parse")

            return "Limit %s" % lim_id

        def fake_limit_update(db, limits_key, lims):
            self.assertEqual(limits_key, 'limits')
            self.lims = lims

        def fake_command(db, control_channel, command, *params):
            self.assertEqual(control_channel, 'control')
            self.cmds.append((command, params))

        self.stubs.Set(tools, 'parse_limit_node', fake_parse_limit_node)
        self.stubs.Set(FakeDatabase, 'limit_update', fake_limit_update)
        self.stubs.Set(FakeDatabase, 'command', fake_command)

    def test_basic(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ["Limit 0", "Limit 1"])
        self.assertEqual(self.cmds, [("reload", ())])
        self.assertEqual(self.stderr.getvalue(), '')

    def test_basic_debug(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, debug=True)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ["Limit 0", "Limit 1"])
        self.assertEqual(self.cmds, [("reload", ())])
        self.assertEqual(self.stderr.getvalue(),
                         "Installing the following limits:\n"
                         "  'Limit 0'\n"
                         "  'Limit 1'\n"
                         "Issuing command: reload\n")

    def test_basic_dryrun(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, dry_run=True)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, None)
        self.assertEqual(self.cmds, [])
        self.assertEqual(self.stderr.getvalue(),
                         "Installing the following limits:\n"
                         "  'Limit 0'\n"
                         "  'Limit 1'\n"
                         "Issuing command: reload\n")

    def test_reload_false(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, do_reload=False)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ['Limit 0', 'Limit 1'])
        self.assertEqual(self.cmds, [])
        self.assertEqual(self.stderr.getvalue(), "")

    def test_reload_string(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, do_reload='spam')

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ['Limit 0', 'Limit 1'])
        self.assertEqual(self.cmds, [('reload', ('spam',))])
        self.assertEqual(self.stderr.getvalue(), "")

    def test_reload_int(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, do_reload=5)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ['Limit 0', 'Limit 1'])
        self.assertEqual(self.cmds, [('reload', ('spread', 5))])
        self.assertEqual(self.stderr.getvalue(), "")

    def test_reload_float(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file, do_reload=3.14)

            self.assertEqual(len(w), 0)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1")])
        self.assertEqual(self.lims, ['Limit 0', 'Limit 1'])
        self.assertEqual(self.cmds, [('reload', ('spread', 3.14))])
        self.assertEqual(self.stderr.getvalue(), "")

    def test_bad_tag(self):
        limits_file = StringIO.StringIO("""<limits>
    <timil id="0"/>
    <limit id="1"/>
</limits>""")

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file)

            self.assertEqual(len(w), 1)
            self.assertIn("Unrecognized tag 'timil' in limits file at index 0",
                          w[-1].message)

        self.assertEqual(self.parsed, [(1, "1")])
        self.assertEqual(self.lims, ['Limit 1'])
        self.assertEqual(self.cmds, [('reload', ())])
        self.assertEqual(self.stderr.getvalue(), "")

    def test_bad_limit(self):
        limits_file = StringIO.StringIO("""<limits>
    <limit id="0"/>
    <limit id="1"/>
    <limit id="2"/>
</limits>""")
        self.fail_idx = 1

        with warnings.catch_warnings(record=True) as w:
            tools._setup_limits('config.file', limits_file)

            self.assertEqual(len(w), 1)
            self.assertIn("Couldn't understand limit at index 1: Failed to "
                          "parse", w[-1].message)

        self.assertEqual(self.parsed, [(0, "0"), (1, "1"), (2, "2")])
        self.assertEqual(self.lims, ['Limit 0', 'Limit 2'])
        self.assertEqual(self.cmds, [('reload', ())])
        self.assertEqual(self.stderr.getvalue(), "")


class TestToolDumpLimits(BaseToolTest):
    def setUp(self):
        super(TestToolDumpLimits, self).setUp()

        self.fakedb._fakedb['limits'] = [
            (10, dict(name="limit1")),
            (20, dict(name="limit2")),
            (30, dict(name="limit3")),
            ]

        self.limits_file = StringIO.StringIO()

        def fake_make_limit_node(root, lim):
            etree.SubElement(root, 'limit', name=lim.name)

        self.stubs.Set(tools, 'make_limit_node', fake_make_limit_node)
        self.stubs.Set(msgpack, 'loads', lambda x: x)
        self.stubs.Set(limits, 'Limit', FakeLimit)

    def test_basic(self):
        tools._dump_limits('config.file', self.limits_file)

        self.assertEqual(self.limits_file.getvalue(),
                         """<?xml version='1.0' encoding='UTF-8'?>
<limits>
  <limit name="limit1"/>
  <limit name="limit2"/>
  <limit name="limit3"/>
</limits>
""")
        self.assertEqual(self.stderr.getvalue(), "")

    def test_debug(self):
        tools._dump_limits('config.file', self.limits_file, debug=True)

        self.assertEqual(self.limits_file.getvalue(),
                         """<?xml version='1.0' encoding='UTF-8'?>
<limits>
  <limit name="limit1"/>
  <limit name="limit2"/>
  <limit name="limit3"/>
</limits>
""")
        self.assertTrue(self.stderr.getvalue().startswith(
                """Dumping limit index 0: {'name': 'limit1'}
Dumping limit index 1: {'name': 'limit2'}
Dumping limit index 2: {'name': 'limit3'}
Dumping limits to file """))



import logging
import unittest

import stubout

from turnstile import utils


LOG = logging.getLogger('turnstile')


class TestHandler(logging.Handler, object):
    def __init__(self):
        super(TestHandler, self).__init__(logging.DEBUG)

        self.messages = []

    def emit(self, record):
        try:
            self.messages.append(self.format(record))
        except Exception:
            pass

    def get_messages(self, clear=False):
        # Get the list of messages and clear it
        messages = self.messages
        if clear:
            self.messages = []
        return messages


# Set up basic logging for tests
test_handler = TestHandler()
LOG.addHandler(test_handler)
LOG.setLevel(logging.DEBUG)
LOG.propagate = False


class TestCase(unittest.TestCase):
    imports = {}

    def setUp(self):
        self.stubs = stubout.StubOutForTesting()

        def fake_import(import_str):
            try:
                return self.imports[import_str]
            except KeyError as exc:
                # Convert into an ImportError
                raise ImportError("Failed to import %s: %s" %
                                  (import_str, exc))

        self.stubs.Set(utils, 'import_class', fake_import)

        # Clear the log messages
        test_handler.get_messages(True)

    def tearDown(self):
        self.stubs.UnsetAll()

        # Clear the log messages
        test_handler.get_messages(True)

    @property
    def log_messages(self):
        # Retrieve and clear test log messages
        return test_handler.get_messages()


class GenericFakeClass(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs



from turnstile import database
from turnstile import middleware

import tests


def preproc1(mid, environ):
    environ.setdefault('turnstile.preprocess', [])
    environ['turnstile.preprocess'].append('preproc1')


def preproc2(mid, environ):
    environ.setdefault('turnstile.preprocess', [])
    environ['turnstile.preprocess'].append('preproc2')


def preproc3(mid, environ):
    environ.setdefault('turnstile.preprocess', [])
    environ['turnstile.preprocess'].append('preproc3')


class FakeMiddleware(tests.GenericFakeClass):
    pass


class FakeMapper(object):
    def __init__(self, delay=None):
        self.delay = delay
        self.environ = None

    def routematch(self, environ):
        self.environ = environ
        if self.delay:
            self.environ['turnstile.delay'] = self.delay


class FakeLimit(object):
    def __init__(self, ident, headers=None):
        self.ident = ident
        self.headers = headers
        self.environ = None
        self.bucket = None
        self.delay = None

    def format(self, status, headers, environ, bucket, delay):
        self.environ = environ
        self.bucket = bucket
        self.delay = delay
        if self.headers:
            headers.update(self.headers)
        return status, "Fake Entity for limit %s" % self.ident


class Response(object):
    def __init__(self):
        self.status = None
        self.headers = None

    def start(self, status, headers):
        self.status = status
        self.headers = dict(headers)


class TestHeadersDict(tests.TestCase):
    def test_init_sequence(self):
        hd = middleware.HeadersDict([('Foo', 'value'), ('bAR', 'VALUE')])

        self.assertEqual(hd.headers, dict(foo='value', bar='VALUE'))

    def test_init_dict(self):
        hd = middleware.HeadersDict(dict(Foo='value', bAR='VALUE'))

        self.assertEqual(hd.headers, dict(foo='value', bar='VALUE'))

    def test_init_kwargs(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        self.assertEqual(hd.headers, dict(foo='value', bar='VALUE'))

    def test_get_item(self):
        hd = middleware.HeadersDict(Foo='value')

        self.assertEqual(hd['foo'], 'value')
        self.assertEqual(hd['Foo'], 'value')
        with self.assertRaises(KeyError):
            foo = hd['bar']

    def test_set_item(self):
        hd = middleware.HeadersDict(Foo='value')

        hd['fOO'] = 'bar'
        self.assertEqual(hd.headers, dict(foo='bar'))
        hd['bAr'] = 'blah'
        self.assertEqual(hd.headers, dict(foo='bar', bar='blah'))

    def test_del_item(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        del hd['fOO']
        self.assertEqual(hd.headers, dict(bar='VALUE'))
        del hd['bar']
        self.assertEqual(hd.headers, {})
        with self.assertRaises(KeyError):
            del hd['baz']

    def test_contains(self):
        hd = middleware.HeadersDict(Foo='value')

        self.assertTrue('foo' in hd)
        self.assertTrue('fOO' in hd)
        self.assertFalse('bAR' in hd)

    def test_iter(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(iter(hd)))
        self.assertEqual(result, ['bar', 'foo'])

    def test_len(self):
        hd = middleware.HeadersDict(Foo='value')

        self.assertEqual(len(hd), 1)
        hd['bAR'] = 'VALUE'
        self.assertEqual(len(hd), 2)

    def test_iterkeys(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.iterkeys()))
        self.assertEqual(result, ['bar', 'foo'])

    def test_iteritems(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.iteritems()))
        self.assertEqual(result, [('bar', 'VALUE'), ('foo', 'value')])

    def test_itervalues(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.itervalues()))
        self.assertEqual(result, ['VALUE', 'value'])

    def test_keys(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.keys()))
        self.assertEqual(result, ['bar', 'foo'])

    def test_items(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.items()))
        self.assertEqual(result, [('bar', 'VALUE'), ('foo', 'value')])

    def test_values(self):
        hd = middleware.HeadersDict(Foo='value', bAR='VALUE')

        result = sorted(list(hd.values()))
        self.assertEqual(result, ['VALUE', 'value'])


class TestTurnstileFilter(tests.TestCase):
    imports = {
        'FakeMiddleware': FakeMiddleware,
        }

    def test_filter_factory(self):
        self.stubs.Set(middleware, 'TurnstileMiddleware',
                       tests.GenericFakeClass)

        result = middleware.turnstile_filter({}, foo='value1', bar='value2')
        self.assertTrue(callable(result))

        obj = result('app')
        self.assertIsInstance(obj, tests.GenericFakeClass)
        self.assertEqual(obj.args, ('app', dict(foo='value1', bar='value2')))

    def test_filter_factory_alternate_middleware(self):
        result = middleware.turnstile_filter({}, foo='value1', bar='value2',
                                             turnstile='FakeMiddleware')
        self.assertTrue(callable(result))

        obj = result('app')
        self.assertIsInstance(obj, FakeMiddleware)
        self.assertEqual(obj.args, ('app', dict(
                    foo='value1',
                    bar='value2',
                    turnstile='FakeMiddleware',
                    )))


class TestTurnstileMiddleware(tests.TestCase):
    imports = {
        'preproc1': preproc1,
        'preproc2': preproc2,
        'preproc3': preproc3,
        }

    def setUp(self):
        super(TestTurnstileMiddleware, self).setUp()
        self.stubs.Set(database, 'initialize', lambda cfg: cfg)
        self.stubs.Set(database, 'ControlDaemon', tests.GenericFakeClass)

    def test_init_basic(self):
        mid = middleware.TurnstileMiddleware('app', {})

        self.assertEqual(mid.app, 'app')
        self.assertEqual(mid.mapper, None)
        self.assertEqual(mid.config, {
                None: dict(status='413 Request Entity Too Large'),
                })
        self.assertEqual(mid.preprocessors, [])
        self.assertEqual(mid.db, {})
        self.assertIsInstance(mid.control_daemon, tests.GenericFakeClass)
        self.assertEqual(mid.control_daemon.args, (mid.db, mid, {}))

    def test_init_config(self):
        config = {
            'status': '404 Not Found',
            'preprocess': 'preproc1 preproc2 preproc3',
            'foo': 'top-level',
            'foo.bar': 'mid-level',
            'foo.bar.baz': 'bottom',
            'redis.host': 'example.com',
            'control.channel': 'spam',
            'control.node_name': 'node1',
            }
        mid = middleware.TurnstileMiddleware('app', config)

        self.assertEqual(mid.app, 'app')
        self.assertEqual(mid.mapper, None)
        self.assertEqual(mid.config, {
                None: dict(
                    status='404 Not Found',
                    preprocess='preproc1 preproc2 preproc3',
                    foo='top-level',
                    ),
                'foo': {
                    'bar': 'mid-level',
                    'bar.baz': 'bottom',
                    },
                'redis': dict(
                    host='example.com',
                    ),
                'control': dict(
                    channel='spam',
                    node_name='node1',
                    ),
                })
        self.assertEqual(mid.preprocessors, [preproc1, preproc2, preproc3])
        self.assertEqual(mid.db, dict(host='example.com'))
        self.assertIsInstance(mid.control_daemon, tests.GenericFakeClass)
        self.assertEqual(mid.control_daemon.args,
                         (mid.db, mid, dict(channel='spam',
                                            node_name='node1')))

    def test_call_through(self):
        response = Response()
        environ = dict(test=True)

        def app(env, start_response):
            self.assertEqual(id(env), id(environ))
            self.assertEqual(start_response, response.start)

            return 'app called'

        mid = middleware.TurnstileMiddleware(app, {})
        mid.mapper = FakeMapper()
        result = mid(environ, response.start)

        self.assertEqual(id(mid.mapper.environ), id(environ))
        self.assertEqual(environ['test'], True)
        self.assertEqual(result, 'app called')
        self.assertEqual(response.status, None)
        self.assertEqual(response.headers, None)

    def test_call_preprocess(self):
        response = Response()
        environ = dict(test=True)

        def app(env, start_response):
            self.assertEqual(id(env), id(environ))
            self.assertEqual(start_response, response.start)

            return 'app called'

        mid = middleware.TurnstileMiddleware(app, dict(
                preprocess='preproc1 preproc2 preproc3',
                ))
        mid.mapper = FakeMapper()
        result = mid(environ, response.start)

        self.assertEqual(id(mid.mapper.environ), id(environ))
        self.assertEqual(environ['turnstile.preprocess'],
                         ['preproc1', 'preproc2', 'preproc3'])
        self.assertEqual(result, 'app called')
        self.assertEqual(response.status, None)
        self.assertEqual(response.headers, None)

    def test_call_limited(self):
        response = Response()
        environ = dict(test=True)
        delays = [
            (1.5, FakeLimit('limit1'), 'bucket1'),
            (3.4, FakeLimit('limit2'), 'bucket2'),
            (2.5, FakeLimit('limit3'), 'bucket3'),
            ]

        def app(env, start_response):
            self.assertTrue(False)

        mid = middleware.TurnstileMiddleware(app, {})
        mid.mapper = FakeMapper(delays)
        result = mid(environ, response.start)

        self.assertEqual(id(mid.mapper.environ), id(environ))
        self.assertEqual(environ['turnstile.delay'], delays)
        self.assertEqual(result, 'Fake Entity for limit limit2')
        self.assertEqual(response.status, '413 Request Entity Too Large')
        self.assertEqual(response.headers, {
                'retry-after': '4',
                })

        self.assertEqual(delays[0][1].environ, None)
        self.assertEqual(delays[0][1].bucket, None)
        self.assertEqual(id(delays[1][1].environ), id(environ))
        self.assertEqual(delays[1][1].bucket, 'bucket2')
        self.assertEqual(delays[2][1].environ, None)
        self.assertEqual(delays[2][1].bucket, None)

    def test_call_limited_alternate_status(self):
        response = Response()
        environ = dict(test=True)
        delays = [(3.4, FakeLimit('limit1'), 'bucket1')]

        def app(env, start_response):
            self.assertTrue(False)

        mid = middleware.TurnstileMiddleware(app, dict(status='404 Not Found'))
        mid.mapper = FakeMapper(delays)
        result = mid(environ, response.start)

        self.assertEqual(id(mid.mapper.environ), id(environ))
        self.assertEqual(environ['turnstile.delay'], delays)
        self.assertEqual(result, 'Fake Entity for limit limit1')
        self.assertEqual(response.status, '404 Not Found')
        self.assertEqual(response.headers, {
                'retry-after': '4',
                })

        self.assertEqual(id(delays[0][1].environ), id(environ))
        self.assertEqual(delays[0][1].bucket, 'bucket1')

    def test_call_limited_extra_headers(self):
        response = Response()
        environ = dict(test=True)
        delays = [(3.4, FakeLimit('limit1', dict(test='foo')), 'bucket1')]

        def app(env, start_response):
            self.assertTrue(False)

        mid = middleware.TurnstileMiddleware(app, {})
        mid.mapper = FakeMapper(delays)
        result = mid(environ, response.start)

        self.assertEqual(id(mid.mapper.environ), id(environ))
        self.assertEqual(environ['turnstile.delay'], delays)
        self.assertEqual(result, 'Fake Entity for limit limit1')
        self.assertEqual(response.status, '413 Request Entity Too Large')
        self.assertEqual(response.headers, {
                'retry-after': '4',
                'test': 'foo',
                })

        self.assertEqual(id(delays[0][1].environ), id(environ))
        self.assertEqual(delays[0][1].bucket, 'bucket1')



from turnstile import utils

import tests


def throw_except(exc_type):
    if exc_type:
        raise exc_type('Thrown')


class TestIgnoreExcept(tests.TestCase):
    def test_ignore_except_no_error(self):
        # Shouldn't raise any exceptions
        finished = False
        with utils.ignore_except():
            throw_except(None)
            finished = True

        self.assertEqual(finished, True)

    def test_ignore_except_with_error(self):
        # Should raise an exception which should be ignored
        finished = False
        with utils.ignore_except():
            throw_except(Exception)
            finished = True

        # Yes, this should be False here...
        self.assertEqual(finished, False)



# Copyright 2012 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ConfigParser
import os
import sys
import warnings

import argparse
from lxml import etree
import msgpack

from turnstile import database
from turnstile import limits
from turnstile import utils


def parse_config(config):
    """
    Parses the connection configuration file.  Establishes a
    connection to the database and returns it, along with the limits
    key and the control channel name, as a tuple.

    :param config: Name of the configuration file, for connecting to
                   the Redis database.
    """

    # Instantiate the config parser
    cp = ConfigParser.SafeConfigParser()

    # Read the limits from the file
    cp.read(config)

    # Make sure we have a connection section
    if not cp.has_section('connection'):
        raise Exception("Missing [connection] section from %r" % config)

    # Get the database configuration...
    db_config = dict(cp.items('connection'))
    limits_key = db_config.pop('limits_key', 'limits')
    control_channel = db_config.pop('control_channel', 'control')

    # Get the database connection
    return database.initialize(db_config), limits_key, control_channel


def parse_limit_node(db, idx, limit):
    """
    Given an XML node describing a limit, return a Limit object.

    :param db: Handle for the Redis database.
    :param idx: The index of the limit in the XML file; used for error
                reporting.
    :param limit: The XML node describing the limit.
    """

    # First, try to import the class; this will raise ImportError if
    # we can't import it
    klass = utils.import_class(limit.get('class'))

    # Build the list of required attributes
    required = set(k for k, v in klass.attrs.items()
                   if 'default' not in v)

    # Now, use introspection on the class to interpret the attributes
    attrs = {}
    for child in limit:
        # Basic validation of child elements
        if child.tag != 'attr':
            warnings.warn("Unrecognized element %r while parsing limit at "
                          "index %d; ignoring..." % (child.tag, idx))
            continue

        # Get the attribute name
        attr = child.get('name')

        # Be liberal in what we accept--ignore unrecognized attributes
        # (with a warning)
        if attr not in klass.attrs:
            warnings.warn("Limit at index %d does not accept an attribute "
                          "%r; ignoring..." % (idx, attr))
            continue

        # OK, get the attribute descriptor
        desc = klass.attrs[attr]

        # Grab the attribute type
        attr_type = desc.get('type', str)

        if attr_type == list:
            # Lists are expressed as child elements; we ignore the
            # child element names
            subtype = desc.get('subtype', str)
            value = []
            for grandchild in child:
                if grandchild.tag != 'value':
                    warnings.warn("Unrecognized element %r while parsing "
                                  "%r attribute of limit at index %d; "
                                  "ignoring..." %
                                  (grandchild.tag, attr, idx))
                    continue

                value.append(subtype(grandchild.text))
        elif attr_type == dict:
            # Dicts are expressed as child elements, with the tags
            # identifying the attribute name
            subtype = desc.get('subtype', str)
            value = {}
            for grandchild in child:
                if grandchild.tag != 'value':
                    warnings.warn("Unrecognized element %r while parsing "
                                  "%r attribute of limit at index %d; "
                                  "ignoring..." %
                                  (grandchild.tag, attr, idx))
                    continue
                elif 'key' not in grandchild.attrib:
                    warnings.warn("Missing 'key' attribute of 'value' "
                                  "element while parsing %r attribute of "
                                  "limit at index %d; ignoring..." %
                                  (attr, idx))
                    continue

                value[grandchild.get('key')] = subtype(grandchild.text)
        else:
            # Simple type conversion
            value = attr_type(child.text)

        # Save the attribute
        attrs[attr] = value

        # Remove from the required set
        required.discard(attr)

    # Did we get all required attributes?
    if required:
        raise TypeError("Missing required attributes %s" %
                        (', '.join(repr(a) for a in sorted(required))))

    # OK, instantiate and return the class
    return klass(db, **attrs)


def _setup_limits(config, limits_file, do_reload=True,
                  dry_run=False, debug=False):
    """
    Set up or update limits in the Redis database.

    :param config: Name of the configuration file, for connecting to
                   the Redis database.
    :param limits_file: Name of the XML file describing the limits to
                        configure.
    :param do_reload: Controls reloading behavior.  If True (the
                      default), a reload command is issued.  If False,
                      no reload command is issued.  String values
                      result in a reload command of the given load
                      type, and integer or float values result in a
                      reload command of type 'spread' with the given
                      spread interval.
    :param dry_run: If True, no changes are made to the database.
                    Implies debug=True.
    :param debug: If True, debugging messages are emitted while
                  loading the limits and updating the database.
    """

    # If dry_run is set, default debug to True
    if dry_run:
        debug = True

    # Connect to the database...
    db, limits_key, control_channel = parse_config(config)

    # Parse the limits file
    limits_tree = etree.parse(limits_file)

    # Now, we parse the limits XML file
    lims = []
    for idx, lim in enumerate(limits_tree.getroot()):
        # Skip tags we don't recognize
        if lim.tag != 'limit':
            warnings.warn("Unrecognized tag %r in limits file at index %d" %
                          (lim.tag, idx))
            continue

        # Construct the limit and add it to the list of limits
        try:
            lims.append(parse_limit_node(db, idx, lim))
        except Exception as exc:
            warnings.warn("Couldn't understand limit at index %d: %s" %
                          (idx, exc))
            continue

    # Now that we have the limits, let's install them
    if debug:
        print >>sys.stderr, "Installing the following limits:"
        for lim in lims:
            print >>sys.stderr, "  %r" % lim
    if not dry_run:
        db.limit_update(limits_key, lims)

    # Were we requested to reload the limits?
    if do_reload is False:
        return

    # OK, figure out what kind of reload to do
    params = []
    if do_reload is True:
        # Nothing to do; use default semantics
        pass
    elif (isinstance(do_reload, (int, long, float)) or
          (isinstance(do_reload, basestring) and do_reload.isdigit())):
        params = ['spread', do_reload]
    else:
        params = [str(do_reload)]

    # Issue the reload command
    if debug:
        cmd = ['reload']
        cmd.extend(params)
        print >>sys.stderr, ("Issuing command: %s" %
                             ' '.join(str(c) for c in cmd))
    if not dry_run:
        db.command(control_channel, 'reload', *params)


def setup_limits():
    """
    Console script entry point for setting up limits from an XML file.
    """

    parser = argparse.ArgumentParser(
        description="Set up or update limits in the Redis database.",
        )

    parser.add_argument('config',
                        help="Name of the configuration file, for connecting "
                        "to the Redis database.")
    parser.add_argument('limits_file',
                        help="Name of the XML file describing the limits to "
                        "configure.")
    parser.add_argument('--debug', '-d',
                        dest='debug',
                        action='store_true',
                        default=False,
                        help="Run the tool in debug mode.")
    parser.add_argument('--dryrun', '--dry_run', '--dry-run', '-n',
                        dest='dry_run',
                        action='store_true',
                        default=False,
                        help="Perform a dry run; inhibits loading data into "
                        "the database.")
    parser.add_argument('--noreload', '-R',
                        dest='reload',
                        action='store_false',
                        default=True,
                        help="Inhibit issuing a reload command.")
    parser.add_argument('--reload-immediate', '-r',
                        dest='reload',
                        action='store_const',
                        const='immediate',
                        help="Cause all nodes to immediately reload the "
                        "limits configuration.")
    parser.add_argument('--reload-spread', '-s',
                        dest='reload',
                        metavar='SECS',
                        type=float,
                        action='store',
                        help="Cause all nodes to reload the limits "
                        "configuration over the specified number of seconds.")

    args = parser.parse_args()
    try:
        _setup_limits(args.config, args.limits_file, args.reload,
                      args.dry_run, args.debug)
    except Exception as exc:
        if args.debug:
            raise
        return str(exc)


def make_limit_node(root, limit):
    """
    Given a Limit object, generate an XML node.

    :param root: The root node of the XML tree being built.
    :param limit: The Limit object to serialize to XML.
    """

    # Build the base limit node
    limit_node = etree.SubElement(root, 'limit',
                                  {'class': limit._limit_full_name})

    # Walk through all the recognized attributes
    for attr in sorted(limit.attrs):
        desc = limit.attrs[attr]
        attr_type = desc.get('type', str)
        value = getattr(limit, attr)

        # Determine the default value, if we have one...
        if 'default' in desc:
            default = (desc['default']() if callable(desc['default']) else
                       desc['default'])

            # Skip attributes that have their default settings
            if value == default:
                continue

        # Set up the attr node
        attr_node = etree.SubElement(limit_node, 'attr', name=attr)

        # Treat lists and dicts specially
        if attr_type == list:
            for val in value:
                val_node = etree.SubElement(attr_node, 'value')
                val_node.text = str(val)
        elif attr_type == dict:
            for key, val in sorted(value.items(), key=lambda x: x[0]):
                val_node = etree.SubElement(attr_node, 'value', key=key)
                val_node.text = str(val)
        else:
            attr_node.text = str(value)


def _dump_limits(config, limits_file, debug=False):
    """
    Dump the current limits from the Redis database.

    :param config: Name of the configuration file, for connecting to
                   the Redis database.
    :param limits_file: Name of the XML file that the limits will be
                        dumped to.
    :param debug: If True, debugging messages are emitted while
                  dumping the limits.
    """

    # Connect to the database...
    db, limits_key, _control_channel = parse_config(config)

    # Now, grab all the limits
    lims = [limits.Limit.hydrate(db, msgpack.loads(lim))
            for lim in db.zrange(limits_key, 0, -1)]

    # Build up the limits tree
    root = etree.Element('limits')
    limit_tree = etree.ElementTree(root)
    for idx, lim in enumerate(lims):
        if debug:
            print >>sys.stderr, "Dumping limit index %d: %r" % (idx, lim)
        make_limit_node(root, lim)

    # Write out the limits file
    if debug:
        print >>sys.stderr, "Dumping limits to file %r" % limits_file
    limit_tree.write(limits_file, xml_declaration=True, encoding='UTF-8',
                     pretty_print=True)


def dump_limits():
    """
    Console script entry point for dumping limits to an XML file.
    """

    parser = argparse.ArgumentParser(
        description="Dump the current limits from the Redis database.",
        )

    parser.add_argument('config',
                        help="Name of the configuration file, for connecting "
                        "to the Redis database.")
    parser.add_argument('limits_file',
                        help="Name of the XML file that the limits will be "
                        "dumped to.")
    parser.add_argument('--debug', '-d',
                        dest='debug',
                        action='store_true',
                        default=False,
                        help="Run the tool in debug mode.")

    args = parser.parse_args()
    try:
        _dump_limits(args.config, args.limits_file, args.debug)
    except Exception as exc:
        if args.debug:
            raise
        return str(exc)



import json
import random

import eventlet
import msgpack
import redis
import routes

from turnstile import database
from turnstile import limits

import tests


class PipelineContext(object):
    def __init__(self, db):
        self._db = db

    def __enter__(self):
        return self._db

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._db._watching = set()


class PubSub(object):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._messages = []
        self._subscriptions = set()

    def subscribe(self, channel):
        self._subscriptions.add(channel)

    def listen(self):
        for msg in self._messages:
            yield msg


class FakeDatabase(database.TurnstileRedis):
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._actions = []
        self._fakedb = {}
        self._expireat = {}
        self._watcherror = {}
        self._watching = set()
        self._messages = []
        self._pubsub = None
        self._published = []

    def pipeline(self):
        self._actions.append(('pipeline',))
        return PipelineContext(self)

    def watch(self, key):
        self._actions.append(('watch', key))
        self._watching.add(key)

    def get(self, key):
        self._actions.append(('get', key))
        return self._fakedb.get(key)

    def multi(self):
        self._actions.append(('multi',))

    def set(self, key, value):
        self._actions.append(('set', key, value))
        self._fakedb[key] = value

    def expireat(self, key, expire):
        self._actions.append(('expireat', key, expire))
        self._expireat[key] = expire

    def execute(self):
        self._actions.append(('execute',))

        for watch in self._watching:
            count = self._watcherror.get(watch, 0)
            if count:
                self._watcherror[watch] = count - 1
                raise redis.WatchError()

    def pubsub(self, *args, **kwargs):
        self._actions.append(('pubsub', args, kwargs))
        self._pubsub = PubSub(*args, **kwargs)
        self._pubsub._messages = self._messages
        return self._pubsub

    def publish(self, channel, msg):
        self._actions.append(('publish', channel, msg))
        self._published.append((channel, msg))

    def zadd(self, key, *args):
        additions = zip(args[::2], args[1::2])
        self._actions.append(('zadd', key, additions))
        zset = self._fakedb.setdefault(key, [])
        for score, item in additions:
            idx = self._zfind(zset, item)
            if idx is None:
                zset.append((score, item))
            else:
                zset[idx] = (score, item)
        zset.sort(key=lambda x: x[0])

    def zrem(self, key, *values):
        self._actions.append(('zrem', key, values))
        zset = self._fakedb.setdefault(key, [])
        removals = []
        for val in values:
            idx = self._zfind(zset, val)
            if idx is not None:
                removals.append(idx)
        for idx in sorted(removals, reverse=True):
            del zset[idx]

    def _zfind(self, zset, value):
        for idx, (score, item) in enumerate(zset):
            if item == value:
                return idx
        return None

    def zrange(self, key, start, stop):
        self._actions.append(('zrange', key, start, stop))
        if key in self._fakedb:
            return [item[1] for item in self._fakedb[key]]
        else:
            return []

    def sadd(self, key, value):
        self._actions.append(('sadd', key, value))
        self._fakedb[key].add(value)

    def execute_command(self, *args, **kwargs):
        self._actions.append(('execute_command', args[0], args[1:], kwargs))
        raise Exception("Unhandled command %s" % args[0])


class ControlDaemonTest(database.ControlDaemon):
    def __init__(self, *args, **kwargs):
        super(ControlDaemonTest, self).__init__(*args, **kwargs)

        self._commands = []

    def _start(self):
        pass

    def _internal(self, *args):
        self._commands.append(('internal', args))

    def test(self, arg):
        self._commands.append(('test', arg))

    def failure(self, *args):
        self._commands.append(('failure', args))
        raise Exception("Failure")


class FakeParser(tests.GenericFakeClass):
    pass


class FakeConnection(object):
    pass


class FakeConnectionPool(tests.GenericFakeClass):
    pass


class FakeLimit(tests.GenericFakeClass):
    @classmethod
    def hydrate(cls, db, limit):
        return cls(db, **limit)

    def dehydrate(self):
        return self.kwargs

    def _route(self, mapper):
        mapper.routes.append(self)


class FakeMapper(tests.GenericFakeClass):
    def __init__(self, *args, **kwargs):
        super(FakeMapper, self).__init__(*args, **kwargs)

        self.routes = []


class FakeFailingMapper(FakeMapper):
    def __init__(self, *args, **kwargs):
        raise Exception("Fake-out")


class ClassTest(object):
    def __init__(self, db, arg1, arg2, a=1, b=2, c=3, d=4, **kwargs):
        self.db = db
        self.args = (arg1, arg2)
        self.a = a
        self.b = b
        self.c = c
        self.d = d

        if 'expire' in kwargs:
            self.expire = kwargs['expire']

        self.hydrated = False

    @classmethod
    def hydrate(cls, db, obj_dict, arg1, arg2):
        obj = cls(db, arg1, arg2, **obj_dict)
        obj.hydrated = True
        return obj

    def dehydrate(self):
        result = {}
        for attr in 'abcd':
            result[attr] = getattr(self, attr)

        if hasattr(self, 'expire'):
            result['expire'] = self.expire

        return result


class TestSafeUpdate(tests.TestCase):
    def setUp(self):
        super(TestSafeUpdate, self).setUp()

        self.callobj = []

        self.stubs.Set(msgpack, 'loads', lambda x: x)
        self.stubs.Set(msgpack, 'dumps', lambda x: x)

    def _update(self, obj):
        self.assertIsInstance(obj, ClassTest)
        self.callobj.append(obj)
        return obj

    def test_make_object(self):
        db = FakeDatabase()
        result = db.safe_update('spam', ClassTest, self._update,
                                'arg1', 'arg2')

        self.assertEqual(len(self.callobj), 1)
        self.assertEqual(self.callobj[-1], result)
        self.assertEqual(result.db, db)
        self.assertEqual(result.args, ('arg1', 'arg2'))
        self.assertEqual(result.a, 1)
        self.assertEqual(result.b, 2)
        self.assertEqual(result.c, 3)
        self.assertEqual(result.d, 4)
        self.assertEqual(result.hydrated, False)
        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'spam'),
                ('get', 'spam'),
                ('multi',),
                ('set', 'spam', dict(a=1, b=2, c=3, d=4)),
                ('execute',),
                ])
        self.assertEqual(db._fakedb, dict(spam=dict(a=1, b=2, c=3, d=4)))
        self.assertEqual(db._expireat, {})

    def test_rehydrate_object(self):
        db = FakeDatabase()
        db._fakedb['spam'] = dict(a=4, b=3, c=2, d=1)
        result = db.safe_update('spam', ClassTest, self._update,
                                'arg1', 'arg2')

        self.assertEqual(len(self.callobj), 1)
        self.assertEqual(self.callobj[-1], result)
        self.assertEqual(result.db, db)
        self.assertEqual(result.args, ('arg1', 'arg2'))
        self.assertEqual(result.a, 4)
        self.assertEqual(result.b, 3)
        self.assertEqual(result.c, 2)
        self.assertEqual(result.d, 1)
        self.assertEqual(result.hydrated, True)
        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'spam'),
                ('get', 'spam'),
                ('multi',),
                ('set', 'spam', dict(a=4, b=3, c=2, d=1)),
                ('execute',),
                ])
        self.assertEqual(db._fakedb, dict(spam=dict(a=4, b=3, c=2, d=1)))
        self.assertEqual(db._expireat, {})

    def test_watch_error(self):
        db = FakeDatabase()
        db._watcherror['spam'] = 1
        result = db.safe_update('spam', ClassTest, self._update,
                                'arg1', 'arg2')

        self.assertEqual(len(self.callobj), 2)
        self.assertEqual(self.callobj[-1], result)
        self.assertEqual(result.db, db)
        self.assertEqual(result.args, ('arg1', 'arg2'))
        self.assertEqual(result.a, 1)
        self.assertEqual(result.b, 2)
        self.assertEqual(result.c, 3)
        self.assertEqual(result.d, 4)
        self.assertEqual(result.hydrated, True)
        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'spam'),
                ('get', 'spam'),
                ('multi',),
                ('set', 'spam', dict(a=1, b=2, c=3, d=4)),
                ('execute',),
                ('watch', 'spam'),
                ('get', 'spam'),
                ('multi',),
                ('set', 'spam', dict(a=1, b=2, c=3, d=4)),
                ('execute',),
                ])
        self.assertEqual(db._fakedb, dict(spam=dict(a=1, b=2, c=3, d=4)))
        self.assertEqual(db._expireat, {})

    def test_expiring_object(self):
        db = FakeDatabase()
        db._fakedb['spam'] = dict(a=4, b=3, c=2, d=1, expire=1000000.0)
        result = db.safe_update('spam', ClassTest, self._update,
                                'arg1', 'arg2')

        self.assertEqual(len(self.callobj), 1)
        self.assertEqual(self.callobj[-1], result)
        self.assertEqual(result.db, db)
        self.assertEqual(result.args, ('arg1', 'arg2'))
        self.assertEqual(result.a, 4)
        self.assertEqual(result.b, 3)
        self.assertEqual(result.c, 2)
        self.assertEqual(result.d, 1)
        self.assertEqual(result.expire, 1000000.0)
        self.assertEqual(result.hydrated, True)
        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'spam'),
                ('get', 'spam'),
                ('multi',),
                ('set', 'spam', dict(a=4, b=3, c=2, d=1, expire=1000000.0)),
                ('expireat', 'spam', 1000000.0),
                ('execute',),
                ])
        self.assertEqual(db._fakedb,
                         dict(spam=dict(a=4, b=3, c=2, d=1, expire=1000000.0)))
        self.assertEqual(db._expireat, dict(spam=1000000.0))


class TestLimitUpdate(tests.TestCase):
    def setUp(self):
        super(TestLimitUpdate, self).setUp()

        self.stubs.Set(msgpack, 'loads', json.loads)
        self.stubs.Set(msgpack, 'dumps', json.dumps)

    def test_limit_update_empty(self):
        db = FakeDatabase()
        limits = [
            FakeLimit(name='limit1'),
            FakeLimit(name='limit2'),
            FakeLimit(name='limit3'),
            ]
        db.limit_update('limits', limits)

        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'limits'),
                ('zrange', 'limits', 0, -1),
                ('multi',),
                ('zadd', 'limits', [(10, '{"name": "limit1"}')]),
                ('zadd', 'limits', [(20, '{"name": "limit2"}')]),
                ('zadd', 'limits', [(30, '{"name": "limit3"}')]),
                ('execute',),
                ])
        self.assertEqual(db._fakedb['limits'], [
                (10, '{"name": "limit1"}'),
                (20, '{"name": "limit2"}'),
                (30, '{"name": "limit3"}'),
                ])

    def test_limit_update_delete(self):
        db = FakeDatabase()
        db._fakedb['limits'] = [
            (10, '{"name": "limit1"}'),
            (20, '{"name": "limit2"}'),
            (30, '{"name": "limit3"}'),
            ]
        db.limit_update('limits', [])

        self.assertEqual(db._actions[:4], [
                ('pipeline',),
                ('watch', 'limits'),
                ('zrange', 'limits', 0, -1),
                ('multi',),
                ])
        self.assertEqual(db._actions[-1], ('execute',))

        tmp = sorted(list(db._actions[4:-1]), key=lambda x: x[2][0])
        self.assertEqual(tmp, [
                ('zrem', 'limits', ('{"name": "limit1"}',)),
                ('zrem', 'limits', ('{"name": "limit2"}',)),
                ('zrem', 'limits', ('{"name": "limit3"}',)),
                ])

        self.assertEqual(db._fakedb['limits'], [])

    def test_limit_update_overlap(self):
        db = FakeDatabase()
        db._fakedb['limits'] = [
            (10, '{"name": "limit1"}'),
            (20, '{"name": "limit3"}'),
            (30, '{"name": "limit4"}'),
            ]
        limits = [
            FakeLimit(name='limit1'),
            FakeLimit(name='limit2'),
            FakeLimit(name='limit3'),
            ]
        db.limit_update('limits', limits)

        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'limits'),
                ('zrange', 'limits', 0, -1),
                ('multi',),
                ('zrem', 'limits', ('{"name": "limit4"}',)),
                ('zadd', 'limits', [(10, '{"name": "limit1"}')]),
                ('zadd', 'limits', [(20, '{"name": "limit2"}')]),
                ('zadd', 'limits', [(30, '{"name": "limit3"}')]),
                ('execute',),
                ])
        self.assertEqual(db._fakedb['limits'], [
                (10, '{"name": "limit1"}'),
                (20, '{"name": "limit2"}'),
                (30, '{"name": "limit3"}'),
                ])

    def test_limit_update_watcherror(self):
        db = FakeDatabase()
        db._fakedb['limits'] = [
            (10, '{"name": "limit1"}'),
            (20, '{"name": "limit3"}'),
            (30, '{"name": "limit4"}'),
            ]
        db._watcherror['limits'] = 1
        limits = [
            FakeLimit(name='limit1'),
            FakeLimit(name='limit2'),
            FakeLimit(name='limit3'),
            ]
        db.limit_update('limits', limits)

        self.assertEqual(db._actions, [
                ('pipeline',),
                ('watch', 'limits'),
                ('zrange', 'limits', 0, -1),
                ('multi',),
                ('zrem', 'limits', ('{"name": "limit4"}',)),
                ('zadd', 'limits', [(10, '{"name": "limit1"}')]),
                ('zadd', 'limits', [(20, '{"name": "limit2"}')]),
                ('zadd', 'limits', [(30, '{"name": "limit3"}')]),
                ('execute',),
                ('watch', 'limits'),
                ('zrange', 'limits', 0, -1),
                ('multi',),
                ('zadd', 'limits', [(10, '{"name": "limit1"}')]),
                ('zadd', 'limits', [(20, '{"name": "limit2"}')]),
                ('zadd', 'limits', [(30, '{"name": "limit3"}')]),
                ('execute',),
                ])
        self.assertEqual(db._fakedb['limits'], [
                (10, '{"name": "limit1"}'),
                (20, '{"name": "limit2"}'),
                (30, '{"name": "limit3"}'),
                ])


class TestCommand(tests.TestCase):
    def test_command_basic(self):
        db = FakeDatabase()
        db.command('control', 'foo')

        self.assertEqual(db._published, [('control', 'foo')])

    def test_command_args(self):
        db = FakeDatabase()
        db.command('control', 'foo', 'bar', 1, 1.0)

        self.assertEqual(db._published, [('control', 'foo:bar:1:1.0')])


class TestInitialize(tests.TestCase):
    imports = {
        'FakeParser': FakeParser,
        'FakeConnection': FakeConnection,
        'FakeConnectionPool': FakeConnectionPool,
        }

    def setUp(self):
        super(TestInitialize, self).setUp()

        self.stubs.Set(database, 'TurnstileRedis', FakeDatabase)

    def test_missing_connection(self):
        with self.assertRaises(redis.ConnectionError):
            db = database.initialize({})

    def test_host_connection(self):
        config = dict(
            host='example.com',
            port='1234',
            db='5',
            password='password',
            socket_timeout='321',
            )
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(db._kwargs, dict(
                host='example.com', port=1234, db=5, password='password',
                socket_timeout=321))

    def test_host_minimal(self):
        config = dict(host='example.com')
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(db._kwargs, dict(host='example.com'))

    def test_unix_connection(self):
        config = dict(
            unix_socket_path='/tmp/redis',
            db='5',
            password='password',
            socket_timeout='321',
            )
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(db._kwargs, dict(
                unix_socket_path='/tmp/redis', db=5, password='password',
                socket_timeout=321))

    def test_unix_minimal(self):
        config = dict(unix_socket_path='/tmp/redis')
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(db._kwargs, dict(unix_socket_path='/tmp/redis'))

    def test_connection_pool_host(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              tests.GenericFakeClass)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=redis.Connection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_unix(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'unix_socket_path': '/tmp/redis',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              tests.GenericFakeClass)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                path='/tmp/redis',
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=redis.UnixDomainSocketConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_host_custom_class(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'connection_pool.connection_class': 'FakeConnection',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              tests.GenericFakeClass)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=FakeConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_unix_custom_class(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'unix_socket_path': '/tmp/redis',
            'connection_pool.connection_class': 'FakeConnection',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              tests.GenericFakeClass)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                unix_socket_path='/tmp/redis',
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=FakeConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_host_custom_pool(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'connection_pool': 'FakeConnectionPool',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              FakeConnectionPool)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=redis.Connection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_unix_custom_pool(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'unix_socket_path': '/tmp/redis',
            'connection_pool': 'FakeConnectionPool',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              FakeConnectionPool)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                path='/tmp/redis',
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=redis.UnixDomainSocketConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_host_custom_pool_custom_class(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'connection_pool': 'FakeConnectionPool',
            'connection_pool.connection_class': 'FakeConnection',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              FakeConnectionPool)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=FakeConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))

    def test_connection_pool_unix_custom_pool_custom_class(self):
        self.stubs.Set(redis, 'ConnectionPool', tests.GenericFakeClass)
        config = {
            'host': 'example.com',
            'port': '1234',
            'db': '5',
            'password': 'password',
            'socket_timeout': '321',
            'unix_socket_path': '/tmp/redis',
            'connection_pool': 'FakeConnectionPool',
            'connection_pool.connection_class': 'FakeConnection',
            'connection_pool.max_connections': '100',
            'connection_pool.parser_class': 'FakeParser',
            'connection_pool.arg1': 'arg1',
            'connection_pool.arg2': '2',
            }
        db = database.initialize(config)

        self.assertEqual(db._args, ())
        self.assertEqual(len(db._kwargs), 1)
        self.assertIn('connection_pool', db._kwargs)
        self.assertIsInstance(db._kwargs['connection_pool'],
                              FakeConnectionPool)

        conn_pool = db._kwargs['connection_pool']
        self.assertEqual(conn_pool.args, ())
        self.assertEqual(conn_pool.kwargs, dict(
                host='example.com',
                port=1234,
                unix_socket_path='/tmp/redis',
                db=5,
                password='password',
                socket_timeout=321,
                connection_class=FakeConnection,
                max_connections=100,
                parser_class=FakeParser,
                arg1='arg1',
                arg2='2',
                ))


class TestControlDaemon(tests.TestCase):
    def setUp(self):
        super(TestControlDaemon, self).setUp()

        # Turn off random number generation
        self.stubs.Set(random, 'random', lambda: 1.0)

    def stub_spawn(self, call=False):
        self.spawns = []

        def fake_spawn_n(method, *args, **kwargs):
            self.spawns.append(('spawn_n', method, args, kwargs))
            if call:
                return method(*args, **kwargs)

        def fake_spawn_after(delay_time, method, *args, **kwargs):
            self.spawns.append(('spawn_after', delay_time, method,
                                args, kwargs))
            if call:
                return method(*args, **kwargs)

        self.stubs.Set(eventlet, 'spawn_n', fake_spawn_n)
        self.stubs.Set(eventlet, 'spawn_after', fake_spawn_after)

    def stub_start(self):
        self.stubs.Set(database.ControlDaemon, '_start', lambda x: None)

    def stub_reload(self, mapper=FakeMapper):
        self.stubs.Set(msgpack, 'loads', lambda x: x)
        self.stubs.Set(limits, 'Limit', FakeLimit)
        self.stubs.Set(routes, 'Mapper', mapper)

    def test_init(self):
        self.stub_spawn(True)

        def fake_reload(obj):
            obj._reloaded = True

        self.stubs.Set(database.ControlDaemon, '_listen', lambda obj: 'listen')
        self.stubs.Set(database.ControlDaemon, '_reload', fake_reload)

        daemon = database.ControlDaemon('db', 'middleware', 'config')

        self.assertEqual(daemon._db, 'db')
        self.assertEqual(daemon._middleware, 'middleware')
        self.assertEqual(daemon._config, 'config')
        self.assertIsInstance(daemon._pending, eventlet.semaphore.Semaphore)
        self.assertEqual(daemon._listen_thread, 'listen')
        self.assertEqual(daemon._reloaded, True)

    def test_listen_basic(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(db._actions, [('pubsub', (), {})])
        self.assertIsInstance(db._pubsub, PubSub)

        pubsub = db._pubsub
        self.assertEqual(pubsub._args, ())
        self.assertEqual(pubsub._kwargs, {})
        self.assertEqual(pubsub._subscriptions, set(['control']))

    def test_listen_shard(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware',
                                        dict(shard_hint='shard'))
        daemon._listen()

        self.assertEqual(db._actions, [('pubsub', (),
                                        dict(shard_hint='shard'))])
        self.assertIsInstance(db._pubsub, PubSub)

        pubsub = db._pubsub
        self.assertEqual(pubsub._args, ())
        self.assertEqual(pubsub._kwargs, dict(shard_hint='shard'))
        self.assertEqual(pubsub._subscriptions, set(['control']))

    def test_listen_control(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware',
                                        dict(channel='spam'))
        daemon._listen()

        self.assertEqual(db._actions, [('pubsub', (), {})])
        self.assertIsInstance(db._pubsub, PubSub)

        pubsub = db._pubsub
        self.assertEqual(pubsub._args, ())
        self.assertEqual(pubsub._kwargs, {})
        self.assertEqual(pubsub._subscriptions, set(['spam']))

    def test_listen_nonmessage(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='nosuch',
                pattern=None,
                channel='control',
                data='test:foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])

    def test_listen_wrongchannel(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='wrongchannel',
                data='test:foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])

    def test_listen_empty(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data=':foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])

    def test_listen_internal(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='_internal:foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])
        self.assertEqual(self.log_messages, [
                "Cannot call internal method '_internal'",
                ])

    def test_listen_unknown(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='unknown:foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])
        self.assertEqual(self.log_messages, [
                "No such command 'unknown'",
                ])

    def test_listen_uncallable(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='spam:foo'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon.spam = 'maps'
        daemon._listen()

        self.assertEqual(daemon._commands, [])
        self.assertEqual(self.log_messages, [
                "Command 'spam' is not callable",
                ])

    def test_listen_badargs(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='test:arg1:arg2'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [])
        self.assertEqual(len(self.log_messages), 1)
        self.assertTrue(self.log_messages[0].startswith(
                "Failed to execute command 'test' arguments "
                "['arg1', 'arg2']"))

    def test_listen_exception(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='failure:arg1:arg2'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [('failure', ('arg1', 'arg2'))])
        self.assertEqual(len(self.log_messages), 1)
        self.assertTrue(self.log_messages[0].startswith(
                "Failed to execute command 'failure' arguments "
                "['arg1', 'arg2']"))

    def test_listen_callout(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='control',
                data='test:arg'))
        daemon = ControlDaemonTest(db, 'middleware', {})
        daemon._listen()

        self.assertEqual(daemon._commands, [('test', 'arg')])
        self.assertEqual(self.log_messages, [])

    def test_listen_callout_alternate_channel(self):
        db = FakeDatabase()
        db._messages.append(dict(
                type='message',
                pattern=None,
                channel='alternate',
                data='test:arg'))
        daemon = ControlDaemonTest(db, 'middleware',
                                   dict(channel='alternate'))
        daemon._listen()

        self.assertEqual(daemon._commands, [('test', 'arg')])
        self.assertEqual(self.log_messages, [])

    def test_ping_nochan(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware', {})
        daemon.ping(None)

        self.assertEqual(db._published, [])

    def test_ping_basic(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware', {})
        daemon.ping('pong')

        self.assertEqual(db._published, [('pong', 'pong')])

    def test_ping_basic_node(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware',
                                        dict(node_name='node'))
        daemon.ping('pong')

        self.assertEqual(db._published, [('pong', 'pong:node')])

    def test_ping_data(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware', {})
        daemon.ping('pong', 'data')

        self.assertEqual(db._published, [('pong', 'pong::data')])

    def test_ping_data_node(self):
        self.stub_start()

        db = FakeDatabase()
        daemon = database.ControlDaemon(db, 'middleware',
                                        dict(node_name='node'))
        daemon.ping('pong', 'data')

        self.assertEqual(db._published, [('pong', 'pong:node:data')])

    def test_reload_command_noargs(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware', {})
        daemon.reload()

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_noargs_configured_bad(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23.5.3'))
        daemon.reload()

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_noargs_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload()

        self.assertEqual(self.spawns, [
                ('spawn_after', 23.0, daemon._reload, (), {})
                ])

    def test_reload_command_badtype(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware', {})
        daemon.reload('badtype')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_badtype_configured_bad(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23.5.3'))
        daemon.reload('badtype')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_badtype_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload('badtype')

        self.assertEqual(self.spawns, [
                ('spawn_after', 23.0, daemon._reload, (), {})
                ])

    def test_reload_command_immediate(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware', {})
        daemon.reload('immediate')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_immediate_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload('immediate')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_spread(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware', {})
        daemon.reload('spread')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_spread_configured_bad(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23.5.3'))
        daemon.reload('spread')

        self.assertEqual(self.spawns, [
                ('spawn_n', daemon._reload, (), {})
                ])

    def test_reload_command_spread_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload('spread')

        self.assertEqual(self.spawns, [
                ('spawn_after', 23.0, daemon._reload, (), {})
                ])

    def test_reload_command_spread_given(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware', {})
        daemon.reload('spread', '18')

        self.assertEqual(self.spawns, [
                ('spawn_after', 18.0, daemon._reload, (), {})
                ])

    def test_reload_command_spread_bad_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload('spread', '18.0.5')

        self.assertEqual(self.spawns, [
                ('spawn_after', 23.0, daemon._reload, (), {})
                ])

    def test_reload_command_spread_given_configured(self):
        self.stub_start()
        self.stub_spawn()

        daemon = database.ControlDaemon('db', 'middleware',
                                        dict(reload_spread='23'))
        daemon.reload('spread', '18')

        self.assertEqual(self.spawns, [
                ('spawn_after', 18.0, daemon._reload, (), {})
                ])

    def test_reload_noacquire(self):
        self.stub_start()
        self.stub_reload()

        db = FakeDatabase()
        db._fakedb['limits'] = [
            (10, dict(limit='limit1')),
            (20, dict(limit='limit2')),
            ]
        middleware = tests.GenericFakeClass()
        daemon = database.ControlDaemon(db, middleware, {})
        daemon._pending.acquire()
        daemon._reload()

        self.assertEqual(db._actions, [])
        self.assertFalse(hasattr(middleware, 'mapper'))

    def test_reload(self):
        self.stub_start()
        self.stub_reload()

        db = FakeDatabase()
        db._fakedb['limits'] = [
            (10, dict(limit='limit1')),
            (20, dict(limit='limit2')),
            ]
        middleware = tests.GenericFakeClass()
        daemon = database.ControlDaemon(db, middleware, {})
        daemon._reload()

        self.assertEqual(db._actions, [('zrange', 'limits', 0, -1)])
        self.assertTrue(hasattr(middleware, 'mapper'))
        self.assertIsInstance(middleware.mapper, FakeMapper)
        self.assertEqual(middleware.mapper.kwargs, dict(register=False))
        self.assertEqual(len(middleware.mapper.routes), 2)
        for idx, route in enumerate(middleware.mapper.routes):
            self.assertIsInstance(route, FakeLimit)
            self.assertEqual(route.args, (db,))
            self.assertEqual(route.kwargs, dict(limit='limit%d' % (idx + 1)))
        self.assertEqual(daemon._pending.balance, 1)

    def test_reload_alternate(self):
        self.stub_start()
        self.stub_reload()

        db = FakeDatabase()
        db._fakedb['alternate'] = [
            (10, dict(limit='limit1')),
            (20, dict(limit='limit2')),
            ]
        middleware = tests.GenericFakeClass()
        daemon = database.ControlDaemon(db, middleware,
                                        dict(limits_key='alternate'))
        daemon._reload()

        self.assertEqual(db._actions, [('zrange', 'alternate', 0, -1)])
        self.assertTrue(hasattr(middleware, 'mapper'))
        self.assertIsInstance(middleware.mapper, FakeMapper)
        self.assertEqual(middleware.mapper.kwargs, dict(register=False))
        self.assertEqual(len(middleware.mapper.routes), 2)
        for idx, route in enumerate(middleware.mapper.routes):
            self.assertIsInstance(route, FakeLimit)
            self.assertEqual(route.args, (db,))
            self.assertEqual(route.kwargs, dict(limit='limit%d' % (idx + 1)))
        self.assertEqual(daemon._pending.balance, 1)

    def test_reload_failure(self):
        self.stub_start()
        self.stub_reload(FakeFailingMapper)

        db = FakeDatabase()
        db._fakedb['errors'] = set()
        middleware = tests.GenericFakeClass()
        daemon = database.ControlDaemon(db, middleware, {})
        daemon._reload()

        self.assertEqual(len(self.log_messages), 1)
        self.assertTrue(self.log_messages[0].startswith(
                'Could not load limits'))
        self.assertEqual(len(db._actions), 3)
        self.assertEqual(db._actions[0], ('zrange', 'limits', 0, -1))
        self.assertEqual(db._actions[1][0], 'sadd')
        self.assertEqual(db._actions[1][1], 'errors')
        self.assertTrue(db._actions[1][2].startswith(
                'Failed to load limits: '))
        self.assertEqual(db._actions[2][0], 'publish')
        self.assertEqual(db._actions[2][1], 'errors')
        self.assertTrue(db._actions[2][2].startswith(
                'Failed to load limits: '))
        self.assertEqual(daemon._pending.balance, 1)

    def test_reload_failure_alternate(self):
        self.stub_start()
        self.stub_reload(FakeFailingMapper)

        db = FakeDatabase()
        db._fakedb['errors_set'] = set()
        middleware = tests.GenericFakeClass()
        daemon = database.ControlDaemon(db, middleware, dict(
                errors_key='errors_set',
                errors_channel='errors_channel',
                ))
        daemon._reload()

        self.assertEqual(len(self.log_messages), 1)
        self.assertTrue(self.log_messages[0].startswith(
                'Could not load limits'))
        self.assertEqual(len(db._actions), 3)
        self.assertEqual(db._actions[0], ('zrange', 'limits', 0, -1))
        self.assertEqual(db._actions[1][0], 'sadd')
        self.assertEqual(db._actions[1][1], 'errors_set')
        self.assertTrue(db._actions[1][2].startswith(
                'Failed to load limits: '))
        self.assertEqual(db._actions[2][0], 'publish')
        self.assertEqual(db._actions[2][1], 'errors_channel')
        self.assertTrue(db._actions[2][2].startswith(
                'Failed to load limits: '))
        self.assertEqual(daemon._pending.balance, 1)



import time

from turnstile import limits

import tests


class TestUnits(tests.TestCase):
    def test_unit_value(self):
        for unit in ('second', 'seconds', 'secs', 'sec', 's'):
            self.assertEqual(limits.get_unit_value(unit), 1)

        for unit in ('minute', 'minutes', 'mins', 'min', 'm'):
            self.assertEqual(limits.get_unit_value(unit), 60)

        for unit in ('hour', 'hours', 'hrs', 'hr', 'h'):
            self.assertEqual(limits.get_unit_value(unit), 3600)

        for unit in ('day', 'days', 'd'):
            self.assertEqual(limits.get_unit_value(unit), 86400)

        self.assertEqual(limits.get_unit_value('31337'), 31337)
        self.assertEqual(limits.get_unit_value(31337), 31337)

        with self.assertRaises(TypeError):
            val = limits.get_unit_value(3133.7)

        with self.assertRaises(KeyError):
            value = limits.get_unit_value('nosuchunit')

    def test_unit_name(self):
        for unit, value in [('second', 1), ('minute', 60), ('hour', 3600),
                            ('day', 86400)]:
            self.assertEqual(limits.get_unit_name(value), unit)

        self.assertEqual(limits.get_unit_name(31337), '31337')


class FakeLimit(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TestBucket(tests.TestCase):
    def test_init(self):
        bucket = limits.Bucket('db', 'limit', 'key')

        self.assertEqual(bucket.db, 'db')
        self.assertEqual(bucket.limit, 'limit')
        self.assertEqual(bucket.key, 'key')
        self.assertEqual(bucket.last, None)
        self.assertEqual(bucket.next, None)
        self.assertEqual(bucket.level, 0.0)

    def test_hydrate(self):
        bucket_dict = dict(last=time.time() - 3600,
                           next=time.time(), level=0.5)
        bucket = limits.Bucket.hydrate('db', bucket_dict, 'limit', 'key')

        self.assertEqual(bucket.db, 'db')
        self.assertEqual(bucket.limit, 'limit')
        self.assertEqual(bucket.key, 'key')
        self.assertEqual(bucket.last, bucket_dict['last'])
        self.assertEqual(bucket.next, bucket_dict['next'])
        self.assertEqual(bucket.level, bucket_dict['level'])

    def test_dehydrate(self):
        bucket_dict = dict(last=time.time() - 3600,
                           next=time.time(), level=0.5)
        bucket = limits.Bucket('db', 'limit', 'key', **bucket_dict)
        newdict = bucket.dehydrate()

        self.assertEqual(bucket_dict, newdict)

    def test_delay_initial(self):
        self.stubs.Set(time, 'time', lambda: 1000000.0)

        limit = FakeLimit(cost=10.0, unit_value=100)
        bucket = limits.Bucket('db', limit, 'key')
        result = bucket.delay({})

        self.assertEqual(result, None)
        self.assertEqual(bucket.last, 1000000.0)
        self.assertEqual(bucket.next, 1000000.0)
        self.assertEqual(bucket.level, 10.0)

    def test_delay_expired(self):
        self.stubs.Set(time, 'time', lambda: 1000000.0)

        limit = FakeLimit(cost=10.0, unit_value=100)
        bucket = limits.Bucket('db', limit, 'key', last=999990.0, level=10.0)
        result = bucket.delay({})

        self.assertEqual(result, None)
        self.assertEqual(bucket.last, 1000000.0)
        self.assertEqual(bucket.next, 1000000.0)
        self.assertEqual(bucket.level, 10.0)

    def test_delay_overlap(self):
        self.stubs.Set(time, 'time', lambda: 1000000.0)

        limit = FakeLimit(cost=10.0, unit_value=100)
        bucket = limits.Bucket('db', limit, 'key', last=999995.0, level=10.0)
        result = bucket.delay({})

        self.assertEqual(result, None)
        self.assertEqual(bucket.last, 1000000.0)
        self.assertEqual(bucket.next, 1000000.0)
        self.assertEqual(bucket.level, 15.0)

    def test_delay_overlimit(self):
        self.stubs.Set(time, 'time', lambda: 1000000.0)

        limit = FakeLimit(cost=10.0, unit_value=100)
        bucket = limits.Bucket('db', limit, 'key', last=999995.0, level=100.0)
        result = bucket.delay({})

        self.assertEqual(result, 5.0)
        self.assertEqual(bucket.last, 1000000.0)
        self.assertEqual(bucket.next, 1000005.0)
        self.assertEqual(bucket.level, 95.0)

    def test_delay_undereps(self):
        self.stubs.Set(time, 'time', lambda: 1000000.0)

        limit = FakeLimit(cost=10.0, unit_value=100)
        bucket = limits.Bucket('db', limit, 'key', last=999995.0, level=95.1)
        result = bucket.delay({})

        self.assertEqual(result, None)
        self.assertEqual(bucket.last, 1000000.0)
        self.assertEqual(bucket.next, 1000000.0)
        self.assertEqual(bucket.level, 100.1)

    def test_messages_empty(self):
        limit = FakeLimit(unit_value=1.0, value=10)
        bucket = limits.Bucket('db', limit, 'key')

        self.assertEqual(bucket.messages, 10)

    def test_messages_half(self):
        limit = FakeLimit(unit_value=1.0, value=10)
        bucket = limits.Bucket('db', limit, 'key', level=0.5)

        self.assertEqual(bucket.messages, 5)

    def test_messages_full(self):
        limit = FakeLimit(unit_value=1.0, value=10)
        bucket = limits.Bucket('db', limit, 'key', level=1.0)

        self.assertEqual(bucket.messages, 0)

    def test_expire(self):
        bucket = limits.Bucket('db', 'limit', 'key', last=1000000.2,
                               level=5.2)

        self.assertEqual(bucket.expire, 1000006)


class LimitTest1(limits.Limit):
    pass


class LimitTest2(limits.Limit):
    attrs = dict(test_attr=dict(
            desc='Test attribute.',
            type=(str,),
            default=''))
    skip = set(['test_skip'])

    def route(self, uri, route_args):
        route_args['route_add'] = 'LimitTest2'
        return 'mod_%s' % uri

    def filter(self, environ, params, unused):
        if 'defer' in environ:
            raise limits.DeferLimit
        environ['test.filter.unused'] = unused
        params['filter_add'] = 'LimitTest2_direct'
        return dict(additional='LimitTest2_additional')


class FakeMapper(object):
    def __init__(self):
        self.routes = []

    def connect(self, name, uri, **kwargs):
        self.routes.append((name, uri, kwargs))


class FakeBucket(object):
    def __init__(self, delay):
        self._params = None
        self._delay = delay

    def delay(self, params):
        self._params = params
        return self._delay


class FakeDatabase(object):
    def __init__(self, bucket):
        self.update = None
        self.bucket = bucket

    def safe_update(self, key, klass, update, *args):
        self.update = (key, klass, update, args)
        return update(self.bucket)


class TestLimitMeta(tests.TestCase):
    def test_registry(self):
        expected = {
            'turnstile.limits:Limit': limits.Limit,
            'tests.test_limits:LimitTest1': LimitTest1,
            'tests.test_limits:LimitTest2': LimitTest2,
            }

        self.assertEqual(limits.LimitMeta._registry, expected)

    def test_full_name(self):
        self.assertEqual(LimitTest1._limit_full_name,
                         'tests.test_limits:LimitTest1')
        self.assertEqual(LimitTest2._limit_full_name,
                         'tests.test_limits:LimitTest2')

    def test_attrs(self):
        base_attrs = set(['uri', 'value', 'unit', 'verbs', 'requirements',
                          'use', 'continue_scan'])

        self.assertEqual(set(limits.Limit.attrs.keys()), base_attrs)
        self.assertEqual(set(LimitTest1.attrs.keys()), base_attrs)
        self.assertEqual(set(LimitTest2.attrs.keys()),
                         base_attrs | set(['test_attr']))

    def test_skip(self):
        base_skip = set(['limit'])

        self.assertEqual(limits.Limit.skip, base_skip)
        self.assertEqual(LimitTest1.skip, base_skip)
        self.assertEqual(LimitTest2.skip, base_skip | set(['test_skip']))


class TestLimit(tests.TestCase):
    imports = {
        'LimitTest1': LimitTest1,
        'LimitTest2': LimitTest2,
        'FakeLimit': FakeLimit,
        }

    def test_init_default(self):
        limit = limits.Limit('db', uri='uri', value=10, unit='second')

        self.assertEqual(limit.db, 'db')
        self.assertEqual(limit.uri, 'uri')
        self.assertEqual(limit._value, 10)
        self.assertEqual(limit._unit, 1)
        self.assertEqual(limit.verbs, [])
        self.assertEqual(limit.requirements, {})
        self.assertEqual(limit.use, None)
        self.assertEqual(limit.continue_scan, True)

    def test_init_missing_value(self):
        with self.assertRaises(TypeError):
            limit = limits.Limit('db')

    def test_init_bad_value(self):
        with self.assertRaises(ValueError):
            limit = limits.Limit('db', uri='uri', value=0, unit=1)

    def test_init_bad_unit(self):
        with self.assertRaises(ValueError):
            limit = limits.Limit('db', uri='uri', value=10, unit=0)

    def test_init_verbs(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             verbs=['get', 'PUT', 'Head'])

        self.assertEqual(limit.verbs, ['GET', 'PUT', 'HEAD'])

    def test_init_requirements(self):
        expected = dict(foo=r'\..*', bar=r'.\.*')
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             requirements=expected)

        self.assertEqual(limit.requirements, expected)

    def test_init_use(self):
        expected = ['spam', 'ni']
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             use=expected)

        self.assertEqual(limit.use, expected)

    def test_init_continue_scan(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             continue_scan=False)

        self.assertEqual(limit.continue_scan, False)

    def test_hydrate(self):
        expected = dict(
            uri='uri',
            value=10,
            unit='second',
            verbs=['GET', 'PUT'],
            requirements=dict(foo=r'\..*', bar=r'.\.*'),
            use=['baz'],
            continue_scan=False,
            )
        exemplar = dict(limit_class='turnstile.limits:Limit')
        exemplar.update(expected)
        limit = limits.Limit.hydrate('db', exemplar)

        self.assertEqual(limit.__class__, limits.Limit)
        self.assertEqual(limit.db, 'db')
        for key, value in expected.items():
            self.assertEqual(getattr(limit, key), value)

    def test_dehydrate(self):
        exemplar = dict(
            uri='uri',
            value=10,
            unit='second',
            verbs=['GET', 'PUT'],
            requirements=dict(foo=r'\..*', bar=r'.\.*'),
            use=['baz'],
            continue_scan=False,
            )
        expected = dict(limit_class='tests.test_limits:LimitTest1')
        expected.update(exemplar)
        limit = LimitTest1('db', **exemplar)

        self.assertEqual(limit.dehydrate(), expected)

    def test_route_basic(self):
        mapper = FakeMapper()
        limit = limits.Limit('db', uri='uri', value=10, unit=1)
        limit._route(mapper)

        kwargs = dict(conditions=dict(function=limit._filter))
        self.assertEqual(mapper.routes, [(None, 'uri', kwargs)])

    def test_route_verbs(self):
        mapper = FakeMapper()
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             verbs=['get', 'post'])
        limit._route(mapper)

        kwargs = dict(conditions=dict(
                function=limit._filter,
                method=['GET', 'POST'],
                ))
        self.assertEqual(mapper.routes, [(None, 'uri', kwargs)])

    def test_route_requirements(self):
        mapper = FakeMapper()
        limit = limits.Limit('db', uri='uri', value=10, unit=1,
                             requirements=dict(foo=r'\..*', bar=r'.\.*'))
        limit._route(mapper)

        kwargs = dict(
            conditions=dict(function=limit._filter),
            requirements=dict(foo=r'\..*', bar=r'.\.*'),
            )
        self.assertEqual(mapper.routes, [(None, 'uri', kwargs)])

    def test_route_hook(self):
        mapper = FakeMapper()
        limit = LimitTest2('db', uri='uri', value=10, unit=1)
        limit._route(mapper)

        kwargs = dict(
            conditions=dict(function=limit._filter),
            route_add='LimitTest2',
            )
        self.assertEqual(mapper.routes, [(None, 'mod_uri', kwargs)])

    def test_key(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)
        params = dict(a=1, b=2, c=3, d=4, e=5, f=6)
        key = limit.key(params)

        self.assertEqual(key, 'turnstile.limits:Limit/a=1/b=2/c=3/d=4/e=5/f=6')

    def test_filter_basic(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1)
        environ = {}
        params = dict(param='test')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit/param=test'

        self.assertEqual(result, False)
        self.assertEqual(environ, {})
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_use(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1, use=['param2'])
        environ = {}
        params = dict(param1='spam', param2='ni')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit/param2=ni'

        self.assertEqual(result, False)
        self.assertEqual(environ, {})
        self.assertEqual(params, dict(param1='spam', param2='ni'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_use_empty(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1, use=[])
        environ = {}
        params = dict(param1='spam', param2='ni')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit'

        self.assertEqual(result, False)
        self.assertEqual(environ, {})
        self.assertEqual(params, dict(param1='spam', param2='ni'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_defer(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = LimitTest2(db, uri='uri', value=10, unit=1)
        environ = dict(defer=True)
        params = dict(param='test')
        result = limit._filter(environ, params)

        self.assertEqual(result, False)
        self.assertEqual(environ, dict(defer=True))
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update, None)
        self.assertEqual(bucket._params, None)

    def test_filter_hook(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = LimitTest2(db, uri='uri', value=10, unit=1)
        environ = {}
        params = dict(param='test')
        result = limit._filter(environ, params)

        key = ('tests.test_limits:LimitTest2/'
               'filter_add=LimitTest2_direct/param=test')

        self.assertEqual(result, False)
        self.assertEqual(environ, {
                'test.filter.unused': {},
                })
        self.assertEqual(params, dict(
                param='test',
                filter_add='LimitTest2_direct',
                additional='LimitTest2_additional',
                ))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_hook_use(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = LimitTest2(db, uri='uri', value=10, unit=1, use=['param2'])
        environ = {}
        params = dict(param1='spam', param2='ni')
        result = limit._filter(environ, params)

        key = ('tests.test_limits:LimitTest2/'
               'filter_add=LimitTest2_direct/param2=ni')

        self.assertEqual(result, False)
        self.assertEqual(environ, {
                'test.filter.unused': dict(param1='spam'),
                })
        self.assertEqual(params, dict(
                param1='spam',
                param2='ni',
                filter_add='LimitTest2_direct',
                additional='LimitTest2_additional',
                ))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_hook_use_empty(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = LimitTest2(db, uri='uri', value=10, unit=1, use=[])
        environ = {}
        params = dict(param1='spam', param2='ni')
        result = limit._filter(environ, params)

        key = ('tests.test_limits:LimitTest2/'
               'filter_add=LimitTest2_direct')

        self.assertEqual(result, False)
        self.assertEqual(environ, {
                'test.filter.unused': dict(param1='spam', param2='ni'),
                })
        self.assertEqual(params, dict(
                param1='spam',
                param2='ni',
                filter_add='LimitTest2_direct',
                additional='LimitTest2_additional',
                ))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_delay(self):
        bucket = FakeBucket(10)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1)
        environ = {}
        params = dict(param='test')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit/param=test'

        self.assertEqual(result, False)
        self.assertEqual(environ, {
                'turnstile.delay': [(10, limit, bucket)],
                })
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_continue_defer(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = LimitTest2(db, uri='uri', value=10, unit=1,
                           continue_scan=False)
        environ = dict(defer=True)
        params = dict(param='test')
        result = limit._filter(environ, params)

        self.assertEqual(result, False)
        self.assertEqual(environ, dict(defer=True))
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update, None)
        self.assertEqual(bucket._params, None)

    def test_filter_continue_delay(self):
        bucket = FakeBucket(10)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1,
                             continue_scan=False)
        environ = {}
        params = dict(param='test')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit/param=test'

        self.assertEqual(result, True)
        self.assertEqual(environ, {
                'turnstile.delay': [(10, limit, bucket)],
                })
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_filter_continue_no_delay(self):
        bucket = FakeBucket(None)
        db = FakeDatabase(bucket)
        limit = limits.Limit(db, uri='uri', value=10, unit=1,
                             continue_scan=False)
        environ = {}
        params = dict(param='test')
        result = limit._filter(environ, params)

        key = 'turnstile.limits:Limit/param=test'

        self.assertEqual(result, True)
        self.assertEqual(environ, {})
        self.assertEqual(params, dict(param='test'))
        self.assertEqual(db.update[0], key)
        self.assertEqual(db.update[1], limits.Bucket)
        self.assertEqual(db.update[3], (limit, key))
        self.assertEqual(id(bucket._params), id(params))

    def test_format(self):
        expected = ("This request was rate-limited.  Please retry your "
                    "request after 1970-01-12T13:46:40Z.")
        status = '413 Request Entity Too Large'
        limit = limits.Limit('db', uri='uri', value=10, unit=1)
        bucket = limits.Bucket('db', limit, 'key', next=1000000.0)
        headers = {}

        result_status, result_entity = limit.format(status, headers, {},
                                                    bucket, 123)

        self.assertEqual(result_status, status)
        self.assertEqual(result_entity, expected)
        self.assertEqual(headers, {'Content-Type': 'text/plain'})

    def test_value_get(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        self.assertEqual(limit.value, 10)

    def test_value_set(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)
        limit.value = 20

        self.assertEqual(limit._value, 20)

    def test_value_set_zero(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        with self.assertRaises(ValueError):
            limit.value = 0

    def test_value_set_negative(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        with self.assertRaises(ValueError):
            limit.value = -1

    def test_unit_value_get(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        self.assertEqual(limit.unit_value, 1)

    def test_unit_value_set(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)
        limit.unit_value = 60

        self.assertEqual(limit._unit, 60)

    def test_unit_value_set_zero(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        with self.assertRaises(ValueError):
            limit.unit_value = 0

    def test_unit_value_set_negative(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        with self.assertRaises(ValueError):
            limit.unit_value = -1

    def test_unit_get(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        for unit, value in [('second', 1), ('minute', 60), ('hour', 3600),
                            ('day', 86400)]:
            limit._unit = value
            self.assertEqual(limit.unit, unit)

        limit._unit = 31337
        self.assertEqual(limit.unit, '31337')

    def test_unit_set(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        for unit in ('second', 'seconds', 'secs', 'sec', 's'):
            limit.unit = unit
            self.assertEqual(limit._unit, 1)

        for unit in ('minute', 'minutes', 'mins', 'min', 'm'):
            limit.unit = unit
            self.assertEqual(limit._unit, 60)

        for unit in ('hour', 'hours', 'hrs', 'hr', 'h'):
            limit.unit = unit
            self.assertEqual(limit._unit, 3600)

        for unit in ('day', 'days', 'd'):
            limit.unit = unit
            self.assertEqual(limit._unit, 86400)

        limit.unit = '31337'
        self.assertEqual(limit._unit, 31337)

        limit.unit = 31337
        self.assertEqual(limit._unit, 31337)

        with self.assertRaises(TypeError):
            limit.unit = 3133.7

        with self.assertRaises(KeyError):
            limit.unit = 'nosuchunit'

        with self.assertRaises(ValueError):
            limit.unit = '0'

        with self.assertRaises(ValueError):
            limit.unit = -1

    def test_cost(self):
        limit = limits.Limit('db', uri='uri', value=10, unit=1)

        self.assertEqual(limit.cost, 0.1)

        limit.unit = 60
        self.assertEqual(limit.cost, 6.0)



# Copyright 2012 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import random
import traceback

import eventlet
import msgpack
import redis
import routes

from turnstile import limits
from turnstile import utils


LOG = logging.getLogger('turnstile')


class TurnstileRedis(redis.StrictRedis):
    def safe_update(self, key, klass, update, *args):
        """
        Safely creates or updates an object in the database.

        :param key: The key the object should be looked up under.
        :param klass: The Python class corresponding to the object.
        :param update: A callable taking a single argument--the object
                       being updated.  The return value of this
                       callable will become the return value of the
                       safe_update() function.  Note that the callable
                       may be called multiple times for a single
                       update.

        If the object does not currently exist in the database, its
        constructor will be called with the database as the first
        parameter, followed by the remaining positional arguments.  If the
        object does currently exist in the database, its hydrate() class
        method will be called, again with the database as the first
        parameter, followed by the remaining positional arguments,
        followed by a dictionary.

        Once the object is built, the update function is called; its
        return value will be saved and the object will be serialized back
        into the database (the object must have a dehydrate() instance
        method taking no parameters and returning a dictionary).  If the
        object has an 'expire' attribute, the key will be set to expire at
        the time given by 'expire'.

        Note that if the key is updated before processing is complete,
        the object will be reloaded and the update function called
        again.  This ensures that the update is atomic, but means that
        the update function must have no side effects other than
        changes to the object it is passed.
        """

        # Do this in a transaction
        with self.pipeline() as pipe:
            while True:
                try:
                    # Watch for changes to the key
                    pipe.watch(key)

                    # Look up or create the object
                    raw = pipe.get(key)
                    if raw is None:
                        obj = klass(self, *args)
                    else:
                        obj = klass.hydrate(self, msgpack.loads(raw), *args)

                    # Start the transaction...
                    pipe.multi()

                    # Call the processor...
                    result = update(obj)

                    # Save the object back to the database
                    pipe.set(key, msgpack.dumps(obj.dehydrate()))

                    # Set its expiration time
                    try:
                        pipe.expireat(key, obj.expire)
                    except AttributeError:
                        # No expiration time available
                        pass

                    # Execute the transaction
                    pipe.execute()
                except redis.WatchError:
                    # Try again...
                    continue
                else:
                    # We're all done!
                    break

        return result

    def limit_update(self, key, limits):
        """
        Safely updates the list of limits in the database.

        :param key: The key the limits are stored under.
        :param limits: A list or sequence of limit objects, each
                       understanding the dehydrate() method.

        The limits list currently in the database will be atomically
        changed to match the new list.  This is done using the
        pipeline() method.
        """

        # Start by dehydrating all the limits
        desired = [msgpack.dumps(l.dehydrate()) for l in limits]
        desired_set = set(desired)

        # Now, let's update the limits
        with self.pipeline() as pipe:
            while True:
                try:
                    # Watch for changes to the key
                    pipe.watch(key)

                    # Look up the existing limits
                    existing = set(pipe.zrange(key, 0, -1))

                    # Start the transaction...
                    pipe.multi()

                    # Remove limits we no longer have
                    for lim in existing - desired_set:
                        pipe.zrem(key, lim)

                    # Update or add all our desired limits
                    for idx, lim in enumerate(desired):
                        pipe.zadd(key, (idx + 1) * 10, lim)

                    # Execute the transaction
                    pipe.execute()
                except redis.WatchError:
                    # Try again...
                    continue
                else:
                    # We're all done!
                    break

    def command(self, channel, command, *args):
        """
        Utility method to issue a command to all Turnstile instances.

        :param channel: The control channel all Turnstile instances
                        are listening on.
        :param command: The command, as plain text.  Currently, only
                        'reload' and 'ping' are recognized.

        All remaining arguments are treated as arguments for the
        command; they will be stringified and sent along with the
        command to the control channel.  Note that ':' is an illegal
        character in arguments, but no warnings will be issued if it
        is used.
        """

        # Build the command we're sending
        cmd = [command]
        cmd.extend(str(a) for a in args)

        # Send it out
        self.publish(channel, ':'.join(cmd))


def initialize(config):
    """
    Initialize a connection to the Redis database.
    """

    # Extract relevant connection information from the configuration
    kwargs = {}
    for cfg_var, type_ in [('host', str), ('port', int), ('db', int),
                           ('password', str), ('socket_timeout', int),
                           ('unix_socket_path', str)]:
        if cfg_var in config:
            kwargs[cfg_var] = type_(config[cfg_var])

    # Make sure we have at a minimum the hostname
    if 'host' not in kwargs and 'unix_socket_path' not in kwargs:
        raise redis.ConnectionError("No host specified for redis database")

    # Look up the connection pool configuration
    cpool_class = None
    cpool = {}
    for key, value in config.items():
        if key.startswith('connection_pool.'):
            _dummy, _sep, varname = key.partition('.')
            if varname == 'connection_class':
                cpool[varname] = utils.import_class(value)
            elif varname == 'max_connections':
                cpool[varname] = int(value)
            elif varname == 'parser_class':
                cpool[varname] = utils.import_class(value)
            else:
                cpool[varname] = value
    if cpool:
        cpool_class = redis.ConnectionPool

    # Use custom connection pool class if requested...
    if 'connection_pool' in config:
        cpool_class = utils.import_class(config['connection_pool'])

    # If we're using a connection pool, we'll need to pass the keyword
    # arguments to that instead of to redis
    if cpool_class:
        cpool.update(kwargs)

        # Use a custom connection class?
        if 'connection_class' not in cpool:
            if 'unix_socket_path' in cpool:
                if 'host' in cpool:
                    del cpool['host']
                if 'port' in cpool:
                    del cpool['port']

                cpool['path'] = cpool['unix_socket_path']
                del cpool['unix_socket_path']

                cpool['connection_class'] = redis.UnixDomainSocketConnection
            else:
                cpool['connection_class'] = redis.Connection

        # Build the connection pool to use and set up to pass it into
        # the redis constructor...
        kwargs = dict(connection_pool=cpool_class(**cpool))

    # Build and return the database
    return TurnstileRedis(**kwargs)


class ControlDaemon(object):
    """
    A daemon thread which listens for control messages and can reload
    the limit configuration from the database.
    """

    def __init__(self, db, middleware, config):
        """
        Initialize the ControlDaemon.  Starts the listening thread and
        triggers an immediate reload.
        """

        # Save some relevant information
        self._db = db
        self._middleware = middleware
        self._config = config

        # Need a semaphore to cover reloads in action
        self._pending = eventlet.semaphore.Semaphore()

        # Start the daemon
        self._start()

    def _start(self):
        """
        Starts the ControlDaemon by launching the listening thread and
        triggering the initial limits load.  Broken out of __init__()
        for testing.
        """

        # Spawn the listening thread
        self._listen_thread = eventlet.spawn_n(self._listen)

        # Now do the initial load
        self._reload()

    def _listen(self):
        """
        Listen for incoming control messages.

        If the 'redis.shard_hint' configuration is set, its value will
        be passed to the pubsub() method when setting up the
        subscription.  The control channel to subscribe to is
        specified by the 'redis.control_channel' configuration
        ('control' by default).
        """

        # Need a pub-sub object
        kwargs = {}
        if 'shard_hint' in self._config:
            kwargs['shard_hint'] = self._config['shard_hint']
        pubsub = self._db.pubsub(**kwargs)

        # Subscribe to the right channel(s)...
        channel = self._config.get('channel', 'control')
        pubsub.subscribe(channel)

        # Now we listen...
        for msg in pubsub.listen():
            # Only interested in messages to our reload channel
            if (msg['type'] in ('pmessage', 'message') and
                msg['channel'] == channel):
                # Figure out what kind of message this is
                command, _sep, args = msg['data'].partition(':')

                # We must have some command...
                if not command:
                    continue

                # Don't do anything with internal commands
                if command[0] == '_':
                    LOG.error("Cannot call internal method %r" % command)
                    continue

                # Don't do anything with missing commands
                try:
                    method = getattr(self, command)
                except AttributeError:
                    LOG.error("No such command %r" % command)
                    continue

                # Don't do anything with non-callables
                if not callable(method):
                    LOG.error("Command %r is not callable" % command)
                    continue

                # Execute the desired command
                arglist = args.split(':')
                try:
                    method(*arglist)
                except Exception:
                    LOG.exception("Failed to execute command %r arguments %r" %
                                  (command, arglist))
                    continue

    def _reload(self):
        """
        Reloads the limits configuration from the database.

        If an error occurs loading the configuration, an error-level
        log message will be emitted.  Additionally, the error message
        will be added to the set specified by the 'redis.errors_key'
        configuration ('errors' by default) and sent to the publishing
        channel specified by the 'redis.errors_channel' configuration
        ('errors' by default).
        """

        # Acquire the pending semaphore.  If we fail, exit--someone
        # else is already doing the reload
        if not self._pending.acquire(False):
            return

        # Do the remaining steps in a try/finally block so we make
        # sure to release the semaphore
        try:
            # Load all the limits
            key = self._config.get('limits_key', 'limits')
            lims = [limits.Limit.hydrate(self._db, msgpack.loads(lim))
                    for lim in self._db.zrange(key, 0, -1)]

            # Build the routes mapper
            mapper = routes.Mapper(register=False)
            for lim in lims:
                lim._route(mapper)

            # Install it
            self._middleware.limits = lims
            self._middleware.mapper = mapper
        except Exception:
            # Log an error
            LOG.exception("Could not load limits")

            # Get our error set and publish channel
            error_key = self._config.get('errors_key', 'errors')
            error_channel = self._config.get('errors_channel', 'errors')

            # Get an informative message
            msg = "Failed to load limits: " + traceback.format_exc()

            # Store the message into the error set.  We use a set here
            # because it's likely that more than one node will
            # generate the same message if there is an error, and this
            # avoids an explosion in the size of the set.
            with utils.ignore_except():
                self._db.sadd(error_key, msg)

            # Publish the message to a channel
            with utils.ignore_except():
                self._db.publish(error_channel, msg)
        finally:
            self._pending.release()

    def ping(self, channel, data=None):
        """
        Process the 'ping' control message.

        :param channel: The publish channel to which to send the
                        response.
        :param data: Optional extra data.  Will be returned as the
                     second argument of the response.

        Responds to the named channel with a command of 'pong' and
        with the node_name (if configured) and provided data as
        arguments.
        """

        if not channel:
            # No place to reply to
            return

        # Get our configured node name
        node_name = self._config.get('node_name')

        # Format the response
        reply = ['pong']
        if node_name or data:
            reply.append(node_name or '')
        if data:
            reply.append(data)

        # And send it
        with utils.ignore_except():
            self._db.publish(channel, ':'.join(reply))

    def reload(self, load_type=None, spread=None):
        """
        Process the 'reload' control message.

        :param load_type: Optional type of reload.  If given as
                          'immediate', reload is triggered
                          immediately.  If given as 'spread', reload
                          is triggered after a random period of time
                          in the interval (0.0, spread).  Otherwise,
                          reload will be as configured.
        :param spread: Optional argument for 'spread' load_type.  Must
                       be a float giving the maximum length of the
                       interval, in seconds, over which the reload
                       should be scheduled.  If not provided, falls
                       back to configuration.

        If a recognized load_type is not given, or is given as
        'spread' but the spread parameter is not a valid float, the
        configuration will be checked for the 'redis.reload_spread'
        value.  If that is a valid value, the reload will be randomly
        scheduled for some time within the interval (0.0,
        redis.reload_spread).
        """

        # Figure out what type of reload this needs to be
        if load_type == 'immediate':
            spread = None
        elif load_type == 'spread':
            try:
                spread = float(spread)
            except (TypeError, ValueError):
                # Not a valid float; use the configured spread value
                load_type = None
        else:
            load_type = None

        if load_type is None:
            # Use configured set-up; see if we have a spread
            # configured
            try:
                spread = float(self._config['reload_spread'])
            except (TypeError, ValueError, KeyError):
                # No valid configuration
                spread = None

        if spread:
            # Apply a randomization to spread the load around
            eventlet.spawn_after(random.random() * spread, self._reload)
        else:
            # Spawn in immediate mode
            eventlet.spawn_n(self._reload)



# Copyright 2012 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math
import time

import metatools

from turnstile import utils


# Recognized units and their names and aliases
_units_list = [
    (1, ('second', 'seconds', 'secs', 'sec', 's')),
    (60, ('minute', 'minutes', 'mins', 'min', 'm')),
    (60 * 60, ('hour', 'hours', 'hrs', 'hr', 'h')),
    (60 * 60 * 24, ('day', 'days', 'd')),
    ]


# Build up a mapping of units to names and vice versa
_units_map = {}
for secs, names in _units_list:
    _units_map[secs] = names[0]
    for name in names:
        _units_map[name] = secs


def get_unit_value(name):
    """Given a unit's name, return its value."""

    # Numbers map to numbers
    if isinstance(name, (int, long)):
        return name

    # Only accept strings from here on
    if not isinstance(name, basestring):
        raise TypeError('name must be a string.')

    # Again, numbers map to numbers
    if name.isdigit():
        return int(name)

    # Look it up in the units map
    return _units_map[name.lower()]


def get_unit_name(value):
    """Given a unit's value, return its name."""

    # Return name if we have one, otherwise stringify value
    return _units_map.get(value, str(value))


class DeferLimit(Exception):
    """Exception raised if limit should not be considered."""

    pass


class Bucket(object):
    """
    Represent a "bucket."  A bucket tracks the necessary values for
    application of the leaky bucket algorithm under the control of a
    limit specification.
    """

    attrs = set(['last', 'next', 'level'])
    eps = 0.1

    def __init__(self, db, limit, key, last=None, next=None, level=0.0):
        """
        Initialize a bucket.

        :param db: The database the bucket is in.
        :param limit: The limit associated with this bucket.
        :param key: The key under which this bucket should be stored.
        :param last: The timestamp of the last request.
        :param next: The timestamp of the next permissible request.
        :param level: The current water level in the bucket.
        """

        self.db = db
        self.limit = limit
        self.key = key
        self.last = last
        self.next = next
        self.level = level

    @classmethod
    def hydrate(cls, db, bucket, limit, key):
        """
        Given a key and a bucket dict, as generated by dehydrate(),
        generate an appropriate instance of Bucket.
        """

        return cls(db, limit, key, **bucket)

    def dehydrate(self):
        """Return a dict representing this bucket."""

        # Only concerned about very specific attributes
        result = {}
        for attr in self.attrs:
            result[attr] = getattr(self, attr)

        return result

    def delay(self, params):
        """Determine delay until next request."""

        now = time.time()

        # Initialize last...
        if not self.last:
            self.last = now

        # How much has leaked out?
        leaked = now - self.last

        # Update the last message time
        self.last = now

        # Update the water level
        self.level = max(self.level - leaked, 0)

        # Are we too full?
        difference = self.level + self.limit.cost - self.limit.unit_value
        if difference >= self.eps:
            self.next = now + difference
            return difference

        # OK, raise the water level and set next to an appropriate
        # value
        self.level += self.limit.cost
        self.next = now

        return None

    @property
    def messages(self):
        """Return remaining messages before limiting."""

        return int(math.floor(((self.limit.unit_value - self.level) /
                               self.limit.unit_value) * self.limit.value))

    @property
    def expire(self):
        """Return the estimated expiration time of this bucket."""

        # Round up and convert to an int
        return int(math.ceil(self.last + self.level))


class LimitMeta(metatools.MetaClass):
    """
    Metaclass for limits.
    """

    _registry = {}

    def __new__(mcs, name, bases, namespace):
        """
        Generate a new Limit class.  Adds the full class name to the
        namespace, for the benefit of dehydrate().  Also registers the
        class in the registry, for the benefit of hydrate().
        """

        # Build the full name
        full_name = '%s:%s' % (namespace['__module__'], name)

        # Add it to the namespace
        namespace['_limit_full_name'] = full_name

        # Set up attrs and skip
        namespace.setdefault('attrs', {})
        namespace.setdefault('skip', set())
        for base in mcs.iter_bases(bases):
            mcs.inherit_dict(base, namespace, 'attrs')
            mcs.inherit_set(base, namespace, 'skip')

        # Create the class
        cls = super(LimitMeta, mcs).__new__(mcs, name, bases, namespace)

        # Register the class
        if full_name not in mcs._registry:
            mcs._registry[full_name] = cls

        return cls


class Limit(object):
    """
    Basic limit.  This can be used as an absolute rate limit on a
    given endpoint or set of endpoints.  All other limit classes must
    subclass this class.
    """

    __metaclass__ = LimitMeta

    attrs = dict(
        uri=dict(
            desc=('The URI the limit applies to.  This should be in a syntax '
                  'recognized by Routes, i.e., "/constant/{variable}".  Note '
                  'that this URI may be displayed to the user.  Required.'),
            type=str,
            ),
        value=dict(
            desc=('The permissible number of requests per unit time.  '
                  'Required.'),
            type=int,
            ),
        unit=dict(
            desc=('The unit of time over which the "value" is considered.  '
                  'This may be a string, such as "second", or an integer '
                  'number of seconds, expressed as a string.  Required.'),
            type=str,
            ),
        verbs=dict(
            desc=('The HTTP verbs this limit should apply to.  Optional.  If '
                  'not provided, this limit matches any request to the URI; '
                  'otherwise, only the listed methods match.  Takes a list of '
                  'strings.'),
            type=list,
            subtype=str,
            default=lambda: [],  # Make sure we don't use the *same* list
            xform=lambda verbs: [v.upper() for v in verbs],
            ),
        requirements=dict(
            desc=('A mapping of variable names in the URI to regular '
                  'expressions; may be used to further restrict a given '
                  'variable to a particular string.  This could be used '
                  'to differentiate a request to "/resource/{id}" from '
                  'a request to "/resource/edit".  Optional.'),
            type=dict,
            subtype=str,
            default=lambda: {},  # Make sure we don't use the *same* dict
            ),
        use=dict(
            desc=('A list of parameters derived from the URI which should be '
                  'used to construct the bucket key.  By default, all '
                  'parameters are used; this provides a way to restrict the '
                  'set of used parameters.'),
            type=list,
            subtype=str,
            default=None,
            ),
        continue_scan=dict(
            desc=('A boolean which signals whether to consider limits '
                  'following this one in the list.  If True (the '
                  'default), the remaining limits are scanned even if '
                  'this limit matches.  May be set to False to skip '
                  'remaining limits.'),
            type=bool,
            default=True,
            ),
        )

    skip = set(['limit'])

    bucket_class = Bucket

    def __init__(self, db, **kwargs):
        """
        Initialize a new limit.

        :param db: The database the limit object is in.

        For the permissible keyword arguments, see the `attrs`
        dictionary.
        """

        self.db = db

        # Save the various arguments
        missing = set()
        for attr, desc in self.attrs.items():
            # A value is provided
            if attr in kwargs:
                value = kwargs[attr]

                # Run the transformer, if one was specified
                if 'xform' in desc:
                    value = desc['xform'](value)
            elif 'default' in desc:
                # Use the default value; if it's callable, call it
                value = (desc['default']() if callable(desc['default']) else
                         desc['default'])
            else:
                # Missing attribute
                missing.add(attr)
                continue

            # Save the attribute value
            setattr(self, attr, value)

        # Did we get all the required attributes?
        if missing:
            raise TypeError("Missing required attributes: %s" %
                            ', '.join(sorted(missing)))

    def __repr__(self):
        """
        Return a representation of the limit.
        """

        base = [self._limit_full_name]
        for attr in sorted(self.attrs):
            desc = self.attrs[attr]
            attr_type = desc.get('type', str)

            # Treat lists and dicts specially
            if attr_type == list:
                sublist = [repr(v) for v in getattr(self, attr)]
                value = '[%s]' % ', '.join(sublist)
            elif attr_type == dict:
                sublist = ['%s=%r' % (k, v) for k, v in
                           getattr(self, attr).items()]
                value = '{%s}' % ', '.join(sublist)
            else:
                value = repr(getattr(self, attr))

            base.append('%s=%s' % (attr, value))

        return '<%s at 0x%x>' % (' '.join(base), id(self))

    @classmethod
    def hydrate(cls, db, limit):
        """
        Given a limit dict, as generated by dehydrate(), generate an
        appropriate instance of Limit (or a subclass).  If the
        required limit class cannot be found, returns None.
        """

        # Extract the limit name from the keyword arguments
        cls_name = limit.pop('limit_class')

        # Is it in the registry yet?
        if cls_name not in cls._registry:
            try:
                utils.import_class(cls_name)
            except ImportError:
                # If we failed to import, ignore...
                pass

        # Look it up in the registry
        cls = cls._registry.get(cls_name)

        # Instantiate the thing
        return cls(db, **limit) if cls else None

    def dehydrate(self):
        """Return a dict representing this limit."""

        # Only concerned about very specific attributes
        result = dict(limit_class=self._limit_full_name)
        for attr in self.attrs:
            # Using getattr allows the properties to come into play
            result[attr] = getattr(self, attr)

        return result

    def _route(self, mapper):
        """
        Set up the route(s) corresponding to the limit.  This controls
        which limits are checked against the request.

        :param mapper: The routes.Mapper object to add the route to.
        """

        # Build up the keyword arguments to feed to connect()
        kwargs = dict(conditions=dict(function=self._filter))

        # Restrict the verbs
        if self.verbs:
            kwargs['conditions']['method'] = self.verbs

        # Add requirements, if provided
        if self.requirements:
            kwargs['requirements'] = self.requirements

        # Hook to allow subclasses to override arguments to connect()
        uri = self.route(self.uri, kwargs)

        # Create the route
        mapper.connect(None, uri, **kwargs)

    def route(self, uri, route_args):
        """
        Provides a hook by which additional arguments may be added to
        the route.  For most limits, this should not be needed; use
        the filter() method instead.  Should return the URI to
        connect.

        :param uri: The configured URI.  May be returned unchanged.
        :param route_args: A dictionary of keyword arguments that will
                           be passed to routes.Mapper.connect().  This
                           dictionary should be modified in place.
        """

        return uri

    def key(self, params):
        """
        Given a set of parameters describing the request, compute a
        key for accessing the corresponding bucket.

        :param params: A dictionary of parameters describing the
                       request; this is likely based on the dictionary
                       from routes.
        """

        # Build up the key in pieces
        parts = [self._limit_full_name]
        parts.extend('%s=%s' % (k, params[k])
                     for k in sorted(params)
                     if k not in self.skip)
        return '/'.join(parts)

    def _filter(self, environ, params):
        """
        Performs final filtering of the request to determine if this
        limit applies.  Returns False if the limit does not apply or
        if the call should not be limited, or True to apply the limit.
        """

        # If 'use' is set, use only the listed parameters; we'll add
        # the others back later
        unused = {}
        if self.use is not None:
            for key, value in params.items():
                if key not in self.use:
                    unused[key] = value

            # Do this in a separate step so we avoid changing a
            # dictionary during traversal
            for key in unused:
                del params[key]

        # First, we need to set up any additional params required to
        # get the bucket.  If the DeferLimit exception is thrown, no
        # further processing is performed.
        try:
            additional = self.filter(environ, params, unused) or {}
        except DeferLimit:
            return False

        # Compute the bucket key
        key = self.key(params)

        # Update the parameters...
        params.update(unused)
        params.update(additional)

        def process_bucket(bucket):
            # Determine the delay for the message
            delay = bucket.delay(params)

            return (bucket, delay)

        # Perform a safe fetch and update of the bucket
        bucket, delay = self.db.safe_update(key, self.bucket_class,
                                            process_bucket, self, key)

        # If we found a delay, store the particulars in the
        # environment; this will later be sorted and an error message
        # corresponding to the longest delay returned.
        if delay is not None:
            environ.setdefault('turnstile.delay', [])
            environ['turnstile.delay'].append((delay, self, bucket))

        # Should we continue the route scan?
        return not self.continue_scan

    def filter(self, environ, params, unused):
        """
        Performs final route filtering.  Should add additional
        parameters to the `params` dict that should be used when
        looking up the bucket.  Parameters that should be added to
        params, but which should not be used to look up the bucket,
        may be returned as a dictionary.  If this limit should not be
        applied to this request, raise DeferLimit.  Note that
        parameters that have already been filtered out of `params`
        will be present in the `unused` dictionary, and will be
        automatically added back to `params` after generation of the
        key.

        Note that the Turnstile configuration is available in the
        environment under the "turnstile.config" key.
        """

        pass

    def format(self, status, headers, environ, bucket, delay):
        """
        Formats a response entity.  Returns a tuple of the desired
        status code and the formatted entity.  The default status code
        is passed in, as is a dictionary of headers.

        :param status: The default status code.  Should be returned to
                       the caller, or an alternate selected.  The
                       status code should include both the number and
                       the message, separated by a single space.
        :param headers: A dictionary of headers for the response.
                        Should update the 'Content-Type' header at a
                        minimum.
        :param environ: The WSGI environment for the request.
        :param bucket: The bucket containing the data which caused the
                       delay decision to be made.  This can be used to
                       obtain such information as the next time the
                       request can be made.
        :param delay: The number of seconds by which the request
                      should be delayed.
        """

        # This is a default response entity, which can be overridden
        # by limit subclasses.
        entity = ("This request was rate-limited.  "
                  "Please retry your request after %s." %
                  time.strftime("%Y-%m-%dT%H:%M:%SZ",
                                time.gmtime(bucket.next)))
        headers['Content-Type'] = 'text/plain'

        return status, entity

    @property
    def value(self):
        """Retrieve the value for this limit."""

        return self._value

    @value.setter
    def value(self, value):
        """Change the value for this limit."""

        if value <= 0:
            raise ValueError("Limit value must be > 0")

        self._value = value

    @property
    def unit(self):
        """Retrieve the name of the unit used for this limit."""

        return get_unit_name(self._unit)

    @unit.setter
    def unit(self, value):
        """
        Change the unit for this limit to the specified unit.  The new
        value may be specified as an integer, a string the indicating
        number of seconds, or one of the recognized unit names.
        """

        self.unit_value = get_unit_value(value)

    @property
    def unit_value(self):
        """
        Retrieve the unit used for this limit as an integer number of
        seconds.
        """

        return self._unit

    @unit_value.setter
    def unit_value(self, value):
        """
        Change the unit used for this limit to the given number of
        seconds.
        """

        if value <= 0:
            raise ValueError("Unit value must be > 0")

        self._unit = int(value)

    @property
    def cost(self):
        """
        Retrieve the amount by which a request increases the water
        level in the bucket.
        """

        return float(self.unit_value) / float(self.value)
